# -*- coding: utf-8 -*-
# File: preproc.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/1a79d595f7eda9dc9dc8428f4461680ed2222ab6/examples/FasterRCNN/data.py>
"""

from typing import Any, List, Optional, Tuple

import numpy as np

# pylint: disable=import-error
from tensorpack.dataflow.imgaug import AugmentorList, ImageAugmentor

from ....datapoint.convert import box_to_point4, point4_to_box
from ....utils.detection_types import ImageType, JsonDict
from ....utils.logger import log_once
from .common import filter_boxes_inside_shape, np_iou
from .modeling.model_fpn import get_all_anchors_fpn
from .utils.np_box_ops import area as np_area
from .utils.np_box_ops import ioa as np_ioa

# pylint: enable=import-error


class MalformedData(BaseException):
    """
    Exception class for malformed data
    """


def augment(dp: JsonDict, imgaug_list: List[ImageAugmentor], add_mask: bool) -> JsonDict:
    """
    Augment an image according to a list of augmentors.

    :param dp: A dict with "image","gt_boxes","gt_labels"
    :param imgaug_list: List auf augmentors
    :param add_mask: not implemented

    :return: A dict with augmented image and gt_boxes
    """

    image = dp["image"]
    gt_boxes = dp["gt_boxes"]
    if gt_boxes.ndim == 1:
        print("stop")
    augmentations = AugmentorList(imgaug_list)
    tf_ms = augmentations.get_transform(image)
    image_aug = tf_ms.apply_image(image)
    points = box_to_point4(gt_boxes)
    points = tf_ms.apply_coords(points)
    gt_boxes = point4_to_box(points)

    dp["image"] = image_aug
    dp["gt_boxes"] = gt_boxes

    if len(gt_boxes):
        assert np.min(np_area(gt_boxes)) > 0, "some boxes have zero area"

    if add_mask:
        raise NotImplementedError

    return dp


def anchors_and_labels(
    dp: JsonDict,
    anchor_strides: Tuple[int],
    anchor_sizes: Tuple[int],
    anchor_ratios: Tuple[int],
    max_size: int,
    batch_per_image: int,
    front_ground_ratio: float,
    positive_anchor_threshold: float,
    negative_anchor_threshold: float,
    crowd_overlap_threshold: float,
) -> Optional[JsonDict]:
    """
    Generating anchors and labels.

    :param dp: datapoint image
    :param anchor_strides: the stride between the center of neighbored anchors
    :param anchor_sizes: a list of anchor sizes: https://arxiv.org/abs/1506.01497
    :param anchor_ratios: a list of anchor ratios: https://arxiv.org/abs/1506.01497
    :param max_size: the maximum size a image entering the model can have.
    :param batch_per_image: total (across FPN levels) number of anchors that are marked valid
    :param front_ground_ratio: front ground ratio among selected RPN anchors
    :param positive_anchor_threshold: will keep all anchors with an IOU-threshold above this benchmark
    :param negative_anchor_threshold: will filter all anchors with an IOU-threshold below this benchmark
    :param crowd_overlap_threshold: Anchors which overlap with a crowd box (IOA larger than threshold)
                                    will be ignored. Setting this to a value larger than 1.0 will disable the feature
    """

    gt_boxes = dp["gt_boxes"]
    image = dp["image"]

    try:
        dummy_crowd_ind = np.zeros((gt_boxes.shape[0],), dtype="int8")

        multilevel_anchor_inputs = get_multilevel_rpn_anchor_input(
            image,
            gt_boxes,
            dummy_crowd_ind,
            anchor_strides,
            anchor_sizes,
            anchor_ratios,
            max_size,
            batch_per_image,
            front_ground_ratio,
            positive_anchor_threshold,
            negative_anchor_threshold,
            crowd_overlap_threshold,
        )
        for i, (anchor_labels, anchor_boxes) in enumerate(multilevel_anchor_inputs):
            dp[f"anchor_labels_lvl{i + 2}"] = anchor_labels
            dp[f"anchor_boxes_lvl{i + 2}"] = anchor_boxes

    except (MalformedData, IndexError) as err:
        log_once(f"input {dp['file_name']} is filtered for training: {str(err)}")
        return None

    return dp


def get_multilevel_rpn_anchor_input(
    image: np.array,
    boxes: np.array,
    is_crowd: np.array,
    anchor_strides: Tuple[int],
    anchor_sizes: Tuple[int],
    anchor_ratios: Tuple[int],
    max_size: float,
    batch_per_image: int,
    front_ground_ratio: float,
    positive_anchor_threshold: float,
    negative_anchor_threshold: float,
    crowd_overlap_threshold: float,
) -> List[Tuple[Any, Any]]:
    """
    Generates multilevel rpn anchors

    :param image: an image
    :param crowd_overlap_threshold: Anchors which overlap with a crowd box (IOA larger than threshold)
        will be ignored. Setting this to a value larger than 1.0 will disable the feature.
    :param negative_anchor_threshold: will filter all anchors with an IOU-threshold below this benchmark
    :param positive_anchor_threshold: will keep all anchors with an IOU-threshold above this benchmark
    :param front_ground_ratio: front ground ratio among selected RPN anchors
    :param batch_per_image: total (across FPN levels) number of anchors that are marked valid
    :param max_size: the maximum size an image entering the model can have.
    :param anchor_ratios: a list of anchor ratios: https://arxiv.org/abs/1506.01497
    :param anchor_sizes: a list of anchor sizes: https://arxiv.org/abs/1506.01497
    :param is_crowd: a numpy array with is_crowd indices as defined in coco, size: n,
    :param boxes: float box, gt. shouldn't be changed, size: nx4
    :param anchor_strides: the stride between the center of neighbored anchors

    :return:
        [(fm_labels, fm_boxes)]: Returns a tuple for each FPN level. Each tuple contains the anchor labels and target
                                  boxes for each pixel in the feature map.

        fm_labels: fHxfWx NUM_ANCHOR_RATIOS
        fm_boxes: fHxfWx NUM_ANCHOR_RATIOS x4

    """

    boxes = boxes.copy()
    anchors_per_level = get_all_anchors_fpn(
        strides=anchor_strides,
        sizes=anchor_sizes,
        ratios=anchor_ratios,
        max_size=max_size,
    )
    flatten_anchors_per_level = [k.reshape((-1, 4)) for k in anchors_per_level]
    all_anchors_flatten = np.concatenate(flatten_anchors_per_level, axis=0)

    inside_ind, inside_anchors = filter_boxes_inside_shape(all_anchors_flatten, image.shape[:2])

    anchor_labels, anchor_gt_boxes = get_anchor_labels(
        inside_anchors,
        boxes[is_crowd == 0],
        boxes[is_crowd == 1],
        batch_per_image,
        front_ground_ratio,
        positive_anchor_threshold,
        negative_anchor_threshold,
        crowd_overlap_threshold,
    )

    # map back to all_anchors, then split to each level
    num_all_anchors = all_anchors_flatten.shape[0]
    all_labels = -np.ones((num_all_anchors,), dtype="int32")
    all_labels[inside_ind] = anchor_labels
    all_boxes = np.zeros((num_all_anchors, 4), dtype="float32")
    all_boxes[inside_ind] = anchor_gt_boxes

    start = 0
    end = 0
    multilevel_inputs = []
    for level_anchor in anchors_per_level:
        assert level_anchor.shape[2] == len(anchor_ratios)
        anchor_shape = level_anchor.shape[:3]  # fHxfWxNUM_ANCHOR_RATIOS
        num_anchor_this_level = np.prod(anchor_shape)
        end = start + num_anchor_this_level
        multilevel_inputs.append(
            (all_labels[start:end].reshape(anchor_shape), all_boxes[start:end, :].reshape(anchor_shape + (4,)))
        )
        start = end

    assert end == num_all_anchors, f"{end} != {num_all_anchors}"

    return multilevel_inputs


def get_anchor_labels(
    anchors: ImageType,
    gt_boxes: ImageType,
    crowd_boxes: ImageType,
    batch_per_image: int,
    front_ground_ratio: float,
    positive_anchor_threshold: float,
    negative_anchor_threshold: float,
    crowd_overlap_threshold: float,
) -> (ImageType, ImageType):
    """
    Label each anchor as fg/bg/ignore.

    :param batch_per_image: total (across FPN levels) number of anchors that are marked valid
    :param crowd_boxes: Cx4 float
    :param gt_boxes: Bx4 float, non-crowd
    :param anchors: Ax4 float
    :param crowd_overlap_threshold: Anchors which overlap with a crowd box (IOA larger than threshold)
                                    will be ignored. Setting this to a value larger than 1.0 will disable the feature.
    :param negative_anchor_threshold: will filter all anchors with an IOU-threshold below this benchmark
    :param positive_anchor_threshold: will keep all anchors with an IOU-threshold above this benchmark
    :param front_ground_ratio: front ground ratio among selected RPN anchors

    :return:
        anchor_labels: (A,) int. Each element is {-1, 0, 1}
        anchor_boxes: Ax4. Contains the target gt_box for each anchor when the anchor is fg.
    """

    # This function will modify labels and return the filtered indices
    def filter_box_label(labels: np.array, value: np.array, max_num: int) -> np.array:
        curr_inds = np.where(labels == value)[0]
        if len(curr_inds) > max_num:
            disable_inds = np.random.choice(curr_inds, size=(len(curr_inds) - max_num), replace=False)
            labels[disable_inds] = -1  # ignore them
            curr_inds = np.where(labels == value)[0]
        return curr_inds

    number_anchors, number_gt_boxes = len(anchors), len(gt_boxes)
    if number_gt_boxes == 0:
        # No ground truth. All anchors are either background or ignored.
        anchor_labels = np.zeros((number_anchors,), dtype="int32")
        filter_box_label(anchor_labels, 0, batch_per_image)
        return anchor_labels, np.zeros((number_anchors, 4), dtype="float32")

    box_ious = np_iou(anchors, gt_boxes)  # NA x NB
    ious_argmax_per_anchor = box_ious.argmax(axis=1)  # NA,
    ious_max_per_anchor = box_ious.max(axis=1)
    ious_max_per_gt = np.amax(box_ious, axis=0, keepdims=True)  # 1xNB
    # for each gt, find all those anchors (including ties) that has the max ious with it
    anchors_with_max_iou_per_gt = np.where(box_ious == ious_max_per_gt)[0]

    # Setting NA labels: 1--fg 0--bg -1--ignore
    anchor_labels = -np.ones((number_anchors,), dtype="int32")  # NA,

    # the order of setting neg/pos labels matter
    anchor_labels[anchors_with_max_iou_per_gt] = 1
    anchor_labels[ious_max_per_anchor >= positive_anchor_threshold] = 1
    anchor_labels[ious_max_per_anchor < negative_anchor_threshold] = 0

    # label all non-ignore candidate boxes which overlap crowd as ignore
    if crowd_boxes.size > 0:
        cand_inds = np.where(anchor_labels >= 0)[0]
        cand_anchors = anchors[cand_inds]
        ioas = np_ioa(crowd_boxes, cand_anchors)
        overlap_with_crowd = cand_inds[ioas.max(axis=0) > crowd_overlap_threshold]
        anchor_labels[overlap_with_crowd] = -1

    # Subsample fg labels: ignore some fg if fg is too many
    target_num_fg = int(batch_per_image * front_ground_ratio)
    fg_inds = filter_box_label(anchor_labels, 1, target_num_fg)
    # Keep an image even if there is no foreground anchors
    # if len(fg_inds) == 0:
    #     raise MalformedData("No valid foreground for RPN!")

    # Subsample bg labels. num_bg is not allowed to be too many
    old_num_bg = np.sum(anchor_labels == 0)
    if old_num_bg == 0:
        # No valid bg in this image, skip.
        raise MalformedData("No valid background for RPN!")
    target_num_bg = batch_per_image - len(fg_inds)
    filter_box_label(anchor_labels, 0, target_num_bg)  # ignore return values

    # Set anchor boxes: the best gt_box for each fg anchor
    anchor_boxes = np.zeros((number_anchors, 4), dtype="float32")
    fg_boxes = gt_boxes[ious_argmax_per_anchor[fg_inds], :]
    anchor_boxes[fg_inds, :] = fg_boxes
    # assert len(fg_inds) + np.sum(anchor_labels == 0) == self.cfg.RPN.BATCH_PER_IM
    return anchor_labels, anchor_boxes



# -*- coding: utf-8 -*-
# File: pipe.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains pipeline components that can be plugged into each other and predictors that are invoked by their
 respective component.
"""

from .anngen import *
from .base import *
from .cell import *
from .common import *
from .concurrency import *
from .doctectionpipe import *
from .language import *
from .layout import *
from .lm import *
from .refine import *
from .registry import *
from .segment import *
from .text import *
from .transform import *



# -*- coding: utf-8 -*-
# File: test_pubstruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Testing the module mapper.pubstruct
"""

from math import isclose
from typing import Dict
from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.datapoint.annotation import SummaryAnnotation
from deepdoctection.datapoint.box import BoundingBox
from deepdoctection.mapper import pub_to_image
from deepdoctection.utils.detection_types import JsonDict

# from deepdoctection.utils.settings import names
from deepdoctection.utils.settings import CellType, LayoutType, TableType

from .conftest import get_pubtabnet_white_image
from .data import DatapointPubtabnet


@mark.basic
@patch("deepdoctection.mapper.pubstruct.load_image_from_file", MagicMock(side_effect=get_pubtabnet_white_image))
def test_pub_to_image(
    datapoint_pubtabnet: JsonDict,
    categories_name_as_key_pubtabnet: Dict[str, str],
    pubtabnet_results: DatapointPubtabnet,
) -> None:
    """
    testing pub_to_image is mapping correctly
    """

    load_image = True

    # Act
    pub_to_image_mapper = pub_to_image(categories_name_as_key_pubtabnet, load_image, True, False, False, False, False)
    dp = pub_to_image_mapper(datapoint_pubtabnet)
    datapoint = pubtabnet_results
    assert dp is not None
    test_anns = dp.get_annotation(category_names=LayoutType.cell)

    # Assert
    assert isinstance(test_anns, list) and len(test_anns) >= 1
    first_ann = test_anns[0]
    last_ann = test_anns[-1]

    assert len(test_anns) == datapoint.get_number_cell_anns()
    assert dp.width == datapoint.get_width()
    assert dp.height == datapoint.get_height()

    assert first_ann.category_name == datapoint.get_first_ann_category(False)
    assert first_ann.category_id == datapoint.get_first_ann_category(True)
    assert isinstance(first_ann.bounding_box, BoundingBox)
    assert isclose(first_ann.bounding_box.ulx, datapoint.get_first_ann_box().ulx, rel_tol=1e-15)
    assert isclose(first_ann.bounding_box.uly, datapoint.get_first_ann_box().uly, rel_tol=1e-15)
    assert isclose(first_ann.bounding_box.width, datapoint.get_first_ann_box().w, rel_tol=1e-15)
    assert isclose(first_ann.bounding_box.height, datapoint.get_first_ann_box().h, rel_tol=1e-15)
    assert (
        first_ann.get_sub_category(CellType.header).category_name == datapoint.get_first_ann_sub_category_header_name()
    )
    assert (
        first_ann.get_sub_category(CellType.row_number).category_id
        == datapoint.get_first_ann_sub_category_row_number_id()
    )
    assert (
        first_ann.get_sub_category(CellType.column_number).category_id
        == datapoint.get_first_ann_sub_category_col_number_id()
    )
    assert (
        first_ann.get_sub_category(CellType.row_span).category_id == datapoint.get_first_ann_sub_category_row_span_id()
    )
    assert (
        first_ann.get_sub_category(CellType.column_span).category_id
        == datapoint.get_first_ann_sub_category_col_span_id()
    )

    assert last_ann.category_name == datapoint.get_last_ann_category_name()
    assert last_ann.get_sub_category(CellType.header).category_name == datapoint.get_last_ann_sub_category_header_name()
    assert (
        last_ann.get_sub_category(CellType.row_number).category_id
        == datapoint.get_last_ann_sub_category_row_number_id()
    )
    assert (
        last_ann.get_sub_category(CellType.column_number).category_id
        == datapoint.get_last_ann_sub_category_col_number_id()
    )
    assert last_ann.get_sub_category(CellType.row_span).category_id == datapoint.get_last_ann_sub_category_row_span_id()
    assert (
        last_ann.get_sub_category(CellType.column_span).category_id == datapoint.get_last_ann_sub_category_col_span_id()
    )

    summary_ann = dp.summary
    assert isinstance(summary_ann, SummaryAnnotation)
    assert (
        summary_ann.get_sub_category(TableType.number_of_rows).category_id
        == datapoint.get_summary_ann_sub_category_rows_id()
    )
    assert (
        summary_ann.get_sub_category(TableType.number_of_columns).category_id
        == datapoint.get_summary_ann_sub_category_col_id()
    )
    assert (
        summary_ann.get_sub_category(TableType.max_row_span).category_id
        == datapoint.get_summary_ann_sub_category_row_span_id()
    )
    assert (
        summary_ann.get_sub_category(TableType.max_col_span).category_id
        == datapoint.get_summary_ann_sub_category_col_span_id()
    )


@mark.basic
@patch("deepdoctection.mapper.pubstruct.load_image_from_file", MagicMock(side_effect=get_pubtabnet_white_image))
def test_pub_to_image_when_items_are_added(
    datapoint_pubtabnet: JsonDict, categories_name_as_key_pubtabnet: Dict[str, str]
) -> None:
    """
    testing pub_to_image is adding items "ROW" and "COL" correctly
    """

    load_image = False
    # Act
    pub_to_image_mapper = pub_to_image(categories_name_as_key_pubtabnet, load_image, True, True, False, False, False)
    dp = pub_to_image_mapper(datapoint_pubtabnet)
    assert dp is not None

    test_anns = dp.get_annotation(category_names=TableType.item)
    summary_ann = dp.summary
    assert isinstance(summary_ann, SummaryAnnotation)
    assert isinstance(test_anns, list)

    assert len(test_anns) == int(summary_ann.get_sub_category(TableType.number_of_rows).category_id) + int(
        summary_ann.get_sub_category(TableType.number_of_columns).category_id
    )


@mark.basic
@patch("deepdoctection.mapper.pubstruct.load_image_from_file", MagicMock(side_effect=get_pubtabnet_white_image))
def test_pub_to_image_when_dd_pipe_like(
    datapoint_pubtabnet: JsonDict, categories_name_as_key_pubtabnet: Dict[str, str]
) -> None:
    """
    testing pub_to_image generates Image like dd-analyzer
    """

    # Act
    pub_to_image_mapper = pub_to_image(categories_name_as_key_pubtabnet, True, True, True, True, False, False)
    dp = pub_to_image_mapper(datapoint_pubtabnet)
    assert dp is not None
    table_list = dp.get_annotation(category_names=LayoutType.table)
    assert len(table_list) == 1
    table = table_list[0]
    assert table.image is not None



# -*- coding: utf-8 -*-
# File: pubstruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotations in pubtabnet style structure
"""
import itertools
import os
from typing import Dict, Iterable, List, Optional, Sequence, Tuple

import numpy as np

from ..datapoint import BoundingBox, CategoryAnnotation, ContainerAnnotation, ImageAnnotation
from ..datapoint.annotation import SummaryAnnotation
from ..datapoint.convert import convert_pdf_bytes_to_np_array_v2
from ..datapoint.image import Image
from ..utils.detection_types import JsonDict
from ..utils.fs import load_bytes_from_pdf_file, load_image_from_file
from ..utils.settings import CellType, LayoutType, Relationships, TableType, WordType
from ..utils.utils import is_file_extension
from .maputils import MappingContextManager, curry, maybe_get_fake_score

__all__ = ["pub_to_image"]


def _convert_boxes(dp: JsonDict, height: int) -> JsonDict:

    if "bbox" in dp:
        table_box_3 = height - dp["bbox"][1]
        dp["bbox"][1] = height - dp["bbox"][3]
        dp["bbox"][3] = table_box_3

    for cell in dp["html"]["cells"]:
        if "bbox" in cell:
            cell_box_3 = height - cell["bbox"][1]
            cell["bbox"][1] = height - cell["bbox"][3]
            cell["bbox"][3] = cell_box_3

    return dp


def _get_table_annotation(dp: JsonDict, category_id: str) -> ImageAnnotation:

    ulx, uly, lrx, lry = list(map(float, dp["bbox"]))
    bbox = BoundingBox(absolute_coords=True, ulx=ulx, uly=uly, lrx=lrx, lry=lry)
    annotation = ImageAnnotation(category_name=LayoutType.table, bounding_box=bbox, category_id=category_id)
    return annotation


def _cell_token(html: Sequence[str]) -> List[List[int]]:

    index_rows = [i for i, tag in enumerate(html) if tag == "<tr>"]
    index_cells = [i for i, tag in enumerate(html) if tag in ("<td>", ">")]
    index_rows_tmp = [(index_rows[i], index_rows[i + 1]) for i in range(len(index_rows) - 1)]
    index_rows_tmp.append((index_rows[-1], index_cells[-1]))
    index_cells_tmp = [
        [index_cell for index_cell in index_cells if index_rows_tmp[i][0] < index_cell < index_rows_tmp[i][1]]
        for i in range(len(index_rows_tmp))
    ]
    index_cells_tmp[-1].append(index_cells[-1])
    return index_cells_tmp


def _item_spans(html: Sequence[str], index_cells: Sequence[Sequence[int]], item: str) -> List[List[int]]:

    item_spans = [
        [
            int(html[index_cell - 1].replace(item + "=", "").replace('"', ""))
            if (item in html[index_cell - 1] and html[index_cell] == ">")
            else (
                int(html[index_cell - 2].replace(item + "=", "").replace('"', ""))
                if (item in html[index_cell - 2] and html[index_cell] == ">")
                else 1
            )
            for index_cell in index_cell_per_row
        ]
        for index_cell_per_row in index_cells
    ]
    return item_spans


def _end_of_header(html: Sequence[str]) -> int:
    index_cells = [i for i, tag in enumerate(html) if tag in ("<td>", ">")]
    header_in_html = [i for i, tag in enumerate(html) if tag == "</thead>"]
    index_header_end = max(header_in_html)
    index_header = [i for i in index_cells if i < index_header_end]
    if len(index_header):
        last_index_header_cell = max(index_header)
        return index_cells.index(last_index_header_cell) + 1
    return 0


def tile_table(row_spans: Sequence[Sequence[int]], col_spans: Sequence[Sequence[int]]) -> List[List[int]]:
    """
    Tiles a table according the row and column span scheme. A table can be represented as a list of list, where each
    inner list has the same length. Each cell with a cell id can be located according to their row and column spans in
    that scheme

    :param row_spans: A list of list of row spans
    :param col_spans: A list of list of column spans
    :return: A list of list of the tiling of the table, indicating the precise place of each cell.
    """
    number_of_cols = sum(col_spans[0])
    number_of_rows = len(col_spans)
    cell_ids = []
    i = 1
    for row in col_spans:
        cell_id_per_row = []
        for idx, k in enumerate(itertools.count(i)):

            if idx < len(row):
                i += 1
                cell_id_per_row.append(k)
            else:
                break
        cell_ids.append(cell_id_per_row)

    tiling = [[-1] * number_of_cols for _ in range(number_of_rows)]  # initialize placeholders
    table = zip(cell_ids, row_spans, col_spans)

    for row_id, row in enumerate(table):  # type: ignore # pylint: disable=R1702
        for cell in zip(row[0], row[1], row[2]):  # type: ignore
            cell_id = cell[0]
            row_span = cell[1]
            col_span = cell[2]
            if 0 in tiling[row_id]:  # calculate actual position of the cell
                col = tiling[row_id].index(0)
            else:
                col = len(tiling[row_id]) - tiling[row_id].count(-1)
            # tile the cell
            for rs in range(row_span):
                for cs in range(col_span):
                    if rs >= 1 and cs == 0:  # if rowSpan>=2 every row below needs to be filled with trailing 0
                        fill = 0
                        while fill < col:
                            if tiling[row_id + rs][fill] == -1:
                                tiling[row_id + rs][fill] = 0
                            fill += 1
                    tiling[row_id + rs][col + cs] = cell_id

    np.array(tiling, dtype=np.int32)
    return tiling


def _add_items(image: Image, item_type: str, categories_name_as_key: Dict[str, str], pubtables_like: bool) -> Image:

    item_number = CellType.row_number if item_type == LayoutType.row else CellType.column_number
    item_span = CellType.row_span if item_type == LayoutType.row else CellType.column_span

    summary_key = TableType.number_of_rows if item_type == LayoutType.row else TableType.number_of_columns

    number_of_items = 0

    if image.summary is not None:
        category_item = image.summary.get_sub_category(summary_key)
        number_of_items = int(category_item.category_id)

    cells = image.get_annotation(category_names=LayoutType.cell)
    table: ImageAnnotation

    for item_num in range(1, number_of_items + 1):
        cell_item = list(
            filter(
                lambda x: x.get_sub_category(item_number).category_id == str(item_num), cells  # pylint: disable=W0640
            )
        )
        cell_item = list(filter(lambda x: x.get_sub_category(item_span).category_id == "1", cell_item))
        if cell_item:
            ulx = min([cell.bounding_box.ulx for cell in cell_item if isinstance(cell.bounding_box, BoundingBox)])

            uly = min([cell.bounding_box.uly for cell in cell_item if isinstance(cell.bounding_box, BoundingBox)])

            lrx = max([cell.bounding_box.lrx for cell in cell_item if isinstance(cell.bounding_box, BoundingBox)])

            lry = max([cell.bounding_box.lry for cell in cell_item if isinstance(cell.bounding_box, BoundingBox)])

            if pubtables_like:
                tables = image.get_annotation(category_names=LayoutType.table)
                if not tables:
                    raise ValueError("pubtables_like = True requires table")
                table = tables[0]

                if item_type == LayoutType.row:
                    if table.bounding_box:
                        ulx = table.bounding_box.ulx + 1.0
                        lrx = table.bounding_box.lrx - 1.0
                else:
                    if table.bounding_box:
                        uly = table.bounding_box.uly + 1.0
                        lry = table.bounding_box.lry - 1.0

            item_ann = ImageAnnotation(
                category_id=categories_name_as_key[TableType.item],
                category_name=TableType.item,
                bounding_box=BoundingBox(absolute_coords=True, ulx=ulx, uly=uly, lrx=lrx, lry=lry),
            )
            item_sub_ann = CategoryAnnotation(category_name=item_type)
            item_ann.dump_sub_category(TableType.item, item_sub_ann, image.image_id)
            image.dump(item_ann)

    if pubtables_like:  # pubtables_like:
        items = image.get_annotation(category_names=TableType.item)
        item_type_anns = [ann for ann in items if ann.get_sub_category(TableType.item).category_name == item_type]
        item_type_anns.sort(
            key=lambda x: x.bounding_box.cx  # type: ignore
            if item_type == LayoutType.column
            else x.bounding_box.cy  # type: ignore
        )
        if table.bounding_box:
            tmp_item_xy = table.bounding_box.uly + 1.0 if item_type == LayoutType.row else table.bounding_box.ulx + 1.0
        for idx, item in enumerate(item_type_anns):
            with MappingContextManager(
                dp_name=image.file_name,
                filter_level="bounding box",
                image_annotation={"category_name": item.category_name, "annotation_id": item.annotation_id},
            ):

                box = item.bounding_box
                if box:
                    tmp_next_item_xy = 0.0
                    if idx != len(item_type_anns) - 1:
                        next_box = item_type_anns[idx + 1].bounding_box
                        if next_box:
                            tmp_next_item_xy = (
                                (box.lry + next_box.uly) / 2
                                if item_type == LayoutType.row
                                else (box.lrx + next_box.ulx) / 2
                            )
                    else:
                        if table.bounding_box:
                            tmp_next_item_xy = (
                                table.bounding_box.lry - 1.0
                                if item_type == LayoutType.row
                                else table.bounding_box.lrx - 1.0
                            )

                    new_embedding_box = BoundingBox(
                        ulx=box.ulx if item_type == LayoutType.row else tmp_item_xy,
                        uly=tmp_item_xy if item_type == LayoutType.row else box.uly,
                        lrx=box.lrx if item_type == LayoutType.row else tmp_next_item_xy,
                        lry=tmp_next_item_xy if item_type == LayoutType.row else box.lry,
                        absolute_coords=True,
                    )
                    item.bounding_box = new_embedding_box

                    tmp_item_xy = tmp_next_item_xy

    return image


def row_col_cell_ids(tiling: List[List[int]]) -> List[Tuple[int, int, int]]:
    """
    Infers absolute rows and columns for every cell from the tiling of a table.

    :param tiling: A list of list of tiling of a table as returned from the `_tile_table`
    :return: A list of 3-tuples with row number, column number and cell id.
    """
    indices = sorted(
        [(i + 1, j + 1, cell_id) for i, row in enumerate(tiling) for j, cell_id in enumerate(row)], key=lambda x: x[2]
    )
    seen = set()
    rows_col_cell_ids = [(a, b, c) for a, b, c in indices if not (c in seen or seen.add(c))]  # type: ignore

    return rows_col_cell_ids


def embedding_in_image(dp: Image, html: List[str], categories_name_as_key: Dict[str, str]) -> Image:
    """
    Generating an image, that resembles the output of an analyzer. The layout of the image is table spanning
    the full page, i.e. there is one table image annotation. Moreover, the table annotation has an image, with cells
    as image annotations.

    :param dp: Image
    :param html: list with html tags
    :param categories_name_as_key: category dictionary with all possible annotations
    :return: Image
    """
    image = Image(file_name=dp.file_name, location=dp.location, external_id=dp.image_id + "image")
    image.image = dp.image
    image.set_width_height(dp.width, dp.height)
    table_ann = ImageAnnotation(
        category_name=LayoutType.table,
        category_id=categories_name_as_key[LayoutType.table],
        bounding_box=BoundingBox(absolute_coords=True, ulx=0.0, uly=0.0, lrx=dp.width, lry=dp.height),
    )
    image.dump(table_ann)
    image.image_ann_to_image(table_ann.annotation_id, True)

    # will check if header is in category. Will then manipulate html in order to remove header and if necessary body
    # node.
    html.insert(0, "<table>")
    html.append("</table>")
    if CellType.header not in categories_name_as_key:
        html.remove("<thead>")
        html.remove("</thead>")
        if "<tbody>" in html and "</tbody>" in html:
            html.remove("<tbody>")
            html.remove("</tbody>")

    html_ann = ContainerAnnotation(category_name=TableType.html, value=html)
    table_ann.dump_sub_category(TableType.html, html_ann)
    for ann in dp.get_annotation():
        image.dump(ann)
        assert table_ann.image
        table_ann.image.dump(ann)
        table_ann.dump_relationship(Relationships.child, ann.annotation_id)

    return image


def nth_index(iterable: Iterable[str], value: str, n: int) -> Optional[int]:
    """
    Returns the position of the n-th string value in an iterable, e.g. a list

    :param iterable: e.g. list
    :param value: any value
    :param n: n-th value
    :return: position if n-th value exists else None
    """
    matches = (idx for idx, val in enumerate(iterable) if val == value)
    return next(itertools.islice(matches, n - 1, n), None)


def pub_to_image_uncur(  # pylint: disable=R0914
    dp: JsonDict,
    categories_name_as_key: Dict[str, str],
    load_image: bool,
    fake_score: bool,
    rows_and_cols: bool,
    dd_pipe_like: bool,
    is_fintabnet: bool,
    pubtables_like: bool,
) -> Optional[Image]:
    """
    Map a datapoint of annotation structure as given in the Pubtabnet dataset to an Image structure.
    <https://github.com/ibm-aur-nlp/PubTabNet>

    :param dp: A datapoint in serialized coco format.
    :param categories_name_as_key: A dict of categories, e.g. DatasetCategories.get_categories(name_as_key=True)
    :param load_image: If `True` it will load image to `Image.image`
    :param fake_score: If dp does not contain a score, a fake score with uniform random variables in (0,1)
                       will be added.
    :param rows_and_cols: If set to `True`, synthetic "ITEM" ImageAnnotations will be added.  Each item has a
                          sub-category "row_col" that is equal to "ROW" or "COL".
    :param dd_pipe_like: This will generate an image identical to the output of the dd analyzer (e.g. table and words
                         annotations as well as sub annotations and relationships will be generated)
    :param is_fintabnet: Set `True`, if this mapping is used for generating fintabnet datapoints.
    :param pubtables_like: Set `True`, the area covering the table will be tiled with non overlapping rows and columns
                           without leaving empty space
    :return: Image
    """

    if dd_pipe_like:
        assert load_image, load_image

    idx = dp.get("imgid")
    if not idx:
        idx = dp.get("table_id", "")

    with MappingContextManager(str(idx) + " is malformed") as transforming_context:
        html = dp["html"]["structure"]["tokens"]
        index_cells = _cell_token(html)
        col_spans = _item_spans(html, index_cells, "colspan")
        row_spans = _item_spans(html, index_cells, "rowspan")
        number_of_rows = len(col_spans)
        number_of_cols = sum(col_spans[0])
        _has_header = html[0] == "<thead>"

        if _has_header:
            end_of_header = _end_of_header(html)

        tiling = tile_table(row_spans, col_spans)

        rows_cols_cell_ids = row_col_cell_ids(tiling)
        number_of_cells = len(rows_cols_cell_ids)
        col_spans = list(itertools.chain(*col_spans))  # type: ignore
        row_spans = list(itertools.chain(*row_spans))  # type: ignore

    if transforming_context.context_error:
        return None

    with MappingContextManager(str(idx)) as mapping_context:

        max_rs, max_cs = 0, 0
        if idx is None:
            raise ValueError("No valid datapoint external id")

        image = Image(file_name=os.path.split(dp["filename"])[1], location=dp["filename"], external_id=idx)

        if is_file_extension(dp["filename"], ".png"):
            np_image = load_image_from_file(dp["filename"])
        if is_file_extension(dp["filename"], ".pdf"):
            pdf_bytes = load_bytes_from_pdf_file(dp["filename"])
            np_image = convert_pdf_bytes_to_np_array_v2(pdf_bytes)
            dp = _convert_boxes(dp, np_image.shape[0])

        if load_image and np_image is not None:
            image.image = np_image
        elif np_image is not None:
            image.set_width_height(np_image.shape[1], np_image.shape[0])

        table_ann: Optional[ImageAnnotation] = None
        if is_fintabnet:  # cannot use for synthetic table ann creation
            table_ann = _get_table_annotation(dp, categories_name_as_key[LayoutType.table])
            image.dump(table_ann)

        for idx, value in enumerate(
            zip(rows_cols_cell_ids[::-1], dp["html"]["cells"][::-1], row_spans[::-1], col_spans[::-1])
        ):
            row_col_cell_id = value[0]
            row_number, col_number, cell_id = row_col_cell_id[0], row_col_cell_id[1], row_col_cell_id[2]
            cell = value[1]
            row_span = value[2]
            col_span = value[3]

            if "bbox" in cell:  # empty cells have no box
                ulx, uly, lrx, lry = list(map(float, cell["bbox"]))
                cell_bounding_box = BoundingBox(absolute_coords=True, ulx=ulx, uly=uly, lrx=lrx, lry=lry)
                cell_ann = ImageAnnotation(
                    category_name=LayoutType.cell,
                    bounding_box=cell_bounding_box,
                    category_id=categories_name_as_key[LayoutType.cell],
                    score=maybe_get_fake_score(fake_score),
                )
                cell_ann.dump_sub_category(
                    CellType.row_number,
                    CategoryAnnotation(category_name=CellType.row_number, category_id=str(row_number)),
                    image.image_id,
                )
                cell_ann.dump_sub_category(
                    CellType.column_number,
                    CategoryAnnotation(category_name=CellType.column_number, category_id=str(col_number)),
                    image.image_id,
                )
                cell_ann.dump_sub_category(
                    CellType.row_span,
                    CategoryAnnotation(category_name=CellType.row_span, category_id=str(row_span)),
                    image.image_id,
                )
                cell_ann.dump_sub_category(
                    CellType.column_span,
                    CategoryAnnotation(category_name=CellType.column_span, category_id=str(col_span)),
                    image.image_id,
                )
                if (
                    int(cell_ann.get_sub_category(CellType.row_span).category_id) > 1
                    or int(cell_ann.get_sub_category(CellType.column_span).category_id) > 1
                ):
                    cell_ann.dump_sub_category(
                        CellType.spanning,
                        CategoryAnnotation(category_name=CellType.spanning),
                        image.image_id,
                    )
                else:
                    cell_ann.dump_sub_category(
                        CellType.spanning,
                        CategoryAnnotation(category_name=LayoutType.cell),
                        image.image_id,
                    )

                max_rs = max(max_rs, row_span)  # type: ignore
                max_cs = max(max_cs, col_span)  # type: ignore

                if _has_header:
                    category_name = CellType.header if cell_id <= end_of_header else CellType.body
                    cell_ann.dump_sub_category(
                        CellType.header, CategoryAnnotation(category_name=category_name), image.image_id
                    )
                image.dump(cell_ann)
                if table_ann is not None:
                    table_ann.dump_relationship(Relationships.child, cell_ann.annotation_id)

                if dd_pipe_like:
                    tokens = cell["tokens"]
                    if "<b>" in tokens and "</b>" in tokens:
                        tokens.remove("<b>")
                        tokens.remove("</b>")
                    text = "".join(tokens)
                    # we are not separating each word but view the full table content as one word
                    word = ImageAnnotation(
                        category_name=LayoutType.word,
                        category_id=categories_name_as_key[LayoutType.word],
                        bounding_box=cell_bounding_box,
                    )
                    text_container = ContainerAnnotation(category_name=WordType.characters, value=text)
                    word.dump_sub_category(WordType.characters, text_container)
                    reading_order = CategoryAnnotation(category_name=Relationships.reading_order, category_id="1")
                    word.dump_sub_category(Relationships.reading_order, reading_order)
                    image.dump(word)
                    cell_ann.dump_relationship(Relationships.child, word.annotation_id)

                    index = nth_index(html, "<td>", number_of_cells - idx)
                    if index:
                        html.insert(index + 1, cell_ann.annotation_id)

        summary_ann = SummaryAnnotation(external_id=image.image_id + "SUMMARY")
        summary_ann.dump_sub_category(
            TableType.number_of_rows,
            CategoryAnnotation(category_name=TableType.number_of_rows, category_id=str(number_of_rows)),
            image.image_id,
        )
        summary_ann.dump_sub_category(
            TableType.number_of_columns,
            CategoryAnnotation(category_name=TableType.number_of_columns, category_id=str(number_of_cols)),
            image.image_id,
        )
        summary_ann.dump_sub_category(
            TableType.max_row_span,
            CategoryAnnotation(category_name=TableType.max_row_span, category_id=str(max_rs)),
            image.image_id,
        )
        summary_ann.dump_sub_category(
            TableType.max_col_span,
            CategoryAnnotation(category_name=TableType.max_col_span, category_id=str(max_cs)),
            image.image_id,
        )
        image.summary = summary_ann

        if rows_and_cols or dd_pipe_like:
            image = _add_items(image, LayoutType.row, categories_name_as_key, pubtables_like)
            image = _add_items(image, LayoutType.column, categories_name_as_key, pubtables_like)

        if dd_pipe_like:
            image = embedding_in_image(image, html, categories_name_as_key)

    if mapping_context.context_error:
        return None
    return image


pub_to_image = curry(pub_to_image_uncur)  # using curry as decorator is not possible as picking will
# fail in multiprocessing



# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module fixtures
"""

from typing import Optional

import numpy as np

from deepdoctection.utils.detection_types import ImageType


def get_white_image(path: str) -> Optional[ImageType]:
    """
    white image
    """
    if path:
        return np.ones((794, 596, 3), dtype=np.int32) * 255  # type: ignore
    return None



# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Fixtures for datapoint package testing
"""
from dataclasses import dataclass

import numpy as np
from pytest import fixture

from deepdoctection.datapoint import BoundingBox, convert_np_array_to_b64
from deepdoctection.utils import get_uuid
from deepdoctection.utils.detection_types import ImageType


@dataclass
class Box:
    """
    Coordinates for bounding box testing
    """

    absolute_coords = True
    ulx = 1.0
    uly = 2.0
    lrx = 2.5
    lry = 4.0
    image_width = 10.0
    image_height = 20

    @property
    def w(self) -> float:
        """
        width
        """
        return self.lrx - self.ulx

    @property
    def h(self) -> float:
        """
        height
        """
        return self.lry - self.uly

    @property
    def cx(self) -> float:
        """
        center x
        """
        return self.ulx + 0.5 * self.w

    @property
    def cy(self) -> float:
        """
        center y
        """
        return self.uly + 0.5 * self.h

    @property
    def area(self) -> float:
        """
        area
        """
        return self.w * self.h

    @property
    def ulx_relative(self) -> float:
        """
        ulx relative coordinate
        """
        return self.ulx / self.image_width

    @property
    def uly_relative(self) -> float:
        """
        uly relative coordinate
        """
        return self.uly / self.image_height

    @property
    def lrx_relative(self) -> float:
        """
        lry relative coordinate
        """
        return self.lrx / self.image_width

    @property
    def lry_relative(self) -> float:
        """
        lry relative coordinate
        """
        return self.lry / self.image_height

    @property
    def w_relative(self) -> float:
        """
        width relative coordinate
        """
        return self.w / self.image_width

    @property
    def h_relative(self) -> float:
        """
        height relative coordinate
        """
        return self.h / self.image_height


@fixture(name="box")
def fixture_box() -> Box:
    """
    Box fixture
    """
    return Box()


@dataclass
class WhiteImage:
    """
    np_array, dummy location and file name and ids for testing
    """

    img = np.ones([4, 6, 3], dtype=np.uint8)
    loc = "/testlocation/test"
    file_name = "test_image.png"
    external_id = "1234"
    uuid = "90c05f37-0000-0000-0000-b84f9d14ff44"

    def get_bounding_box(self) -> BoundingBox:
        """
        BoundingBox
        """
        return BoundingBox(ulx=0.0, uly=0.0, width=self.img.shape[1], height=self.img.shape[0], absolute_coords=True)

    @classmethod
    def get_image_id(cls, type_id: str) -> str:
        """
        image_id

        :param type: Either "d" (default), "n" (no external uuid), "u" (external uuid)
        """
        if type_id == "d":
            return get_uuid(cls.loc + cls.file_name)
        if type_id == "n":
            return get_uuid(cls.external_id)
        return cls.uuid

    def get_image_as_b64_string(self) -> str:
        """
        b64_string image representation
        """
        return convert_np_array_to_b64(self.img)

    def get_image_as_np_array(self) -> ImageType:
        """
        np.array(dtype=np.float32) image representation
        """
        return self.img


@fixture(name="image")
def fixture_image() -> WhiteImage:
    """
    TestWhiteImage
    """
    return WhiteImage()


@dataclass
class CatAnn:
    """
    Category and ids for testing
    """

    category_name = "FOO"
    category_id = "1"
    external_id = "567"
    uuid = "00000000-0000-0000-0000-000000000000"

    @classmethod
    def get_annotation_id(cls, type_id: str) -> str:
        """
        annotation_id, if set externally
        :param type: "n" (no external uuid), "u" (external uuid)
        :return: Annotation as uuid
        """
        if type_id == "n":
            return get_uuid(cls.external_id)
        return cls.uuid


@fixture(name="category_ann")
def fixture_category_ann() -> CatAnn:
    """
    TestCatAnn
    """
    return CatAnn()


class TestPdfPage:
    """
    Pdf as bytes, dummy location and file name for testing
    """

    pdf_bytes = (
        b"%PDF-1.3\n1 0 obj\n<<\n/Type /Pages\n/Count 1\n/Kids [ 3 0 R ]\n>>\nendobj\n2 0 obj\n<<\n/Producer "
        b"(PyPDF2)\n>>\nendobj\n3 0 obj\n<<\n/Type /Page\n/Parent 1 0 R\n/Resources <<\n/Font "
        b"<<\n/F1 5 0 R\n>>\n/ProcSet 6 0 R\n>>\n/MediaBox [ 0 0 612 792 ]\n/Contents 7 0 R\n>>\nendobj\n4 "
        b"0 obj\n<<\n/Type /Catalog\n/Pages 1 0 R\n>>\nendobj\n5 0 obj\n<<\n/Type /Font\n/Subtype /Type1\n/Name"
        b" /F1\n/BaseFont /Helvetica\n/Encoding /WinAnsiEncoding\n>>\nendobj\n6 0 obj\n[ /PDF /Text"
        b" ]\nendobj\n7 0 obj\n<<\n/Length 1074\n>>\nstream\n2 J\r\nBT\r\n0 0 0 rg\r\n/F1 0027 Tf\r\n57.3750"
        b" 722.2800 Td\r\n( A Simple PDF File ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 688.6080 Td\r\n("
        b" This is a small demonstration .pdf file - ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 664.7040"
        b" Td\r\n( just for use in the Virtual Mechanics tutorials. More text. And more ) Tj\r\nET\r\nBT\r\n/F1"
        b" 0010 Tf\r\n69.2500 652.7520 Td\r\n( text. And more text. And more text. And more text. )"
        b" Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 628.8480 Td\r\n( And more text. And more text."
        b" And more text. And more text. And more ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 616.8960"
        b" Td\r\n( text. And more text. Boring, zzzzz. And more text. And more text. And )"
        b" Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 604.9440 Td\r\n( more text. And more text. And more"
        b" text. And more text. And more text. ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 592.9920 Td\r\n("
        b" And more text. And more text. ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 569.0880 Td\r\n( And"
        b" more text. And more text. And more text. And more text. And more ) Tj\r\nET\r\nBT\r\n/F1 0010"
        b" Tf\r\n69.2500 557.1360 Td\r\n( text. And more text. And more text. Even more. Continued on page"
        b" 2 ...) Tj\r\nET\r\n\nendstream\nendobj\nxref\n0 8\n0000000000 65535 f \n0000000009 00000 n"
        b" \n0000000068 00000 n \n0000000108 00000 n \n0000000251 00000 n \n0000000300 00000 n \n0000000407"
        b" 00000 n \n0000000437 00000 n \ntrailer\n<<\n/Size 8\n/Root 4 0 R\n/Info 2 0"
        b" R\n>>\nstartxref\n1563\n%%EOF\n"
    )

    loc = "/testlocation/test"
    file_name = "test_image_0.pdf"
    np_array_shape = (3301, 2550, 3)
    np_array_shape_default = (792, 612, 3)

    def get_image_as_pdf_bytes(self) -> bytes:
        """
        pdf in bytes
        """
        return self.pdf_bytes


@fixture(name="pdf_page")
def fixture_pdf_page() -> TestPdfPage:
    """
    TestPdfPage
    """
    return TestPdfPage()



# -*- coding: utf-8 -*-
# File: test_hflayoutlm.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.hflayoutlm
"""
from typing import List
from unittest.mock import MagicMock, patch

from pytest import mark, raises

from deepdoctection.extern.base import SequenceClassResult, TokenClassResult
from deepdoctection.extern.hflayoutlm import (
    HFLayoutLmSequenceClassifier,
    HFLayoutLmTokenClassifier,
    HFLayoutLmv2SequenceClassifier,
    HFLayoutLmv2TokenClassifier,
    HFLayoutLmv3SequenceClassifier,
    HFLayoutLmv3TokenClassifier,
)
from deepdoctection.utils.detection_types import JsonDict
from deepdoctection.utils.file_utils import pytorch_available
from deepdoctection.utils.settings import BioTag, TokenClasses, get_type

from ..mapper.data import DatapointXfund

if pytorch_available():
    import torch


# pylint: disable=W0613
def get_token_class_results(  # type: ignore
    uuids: List[str],
    input_ids,
    attention_mask,
    token_type_ids,
    boxes,
    tokens,
    model,
    images=None,
) -> List[TokenClassResult]:
    """
    token class result list
    """
    return DatapointXfund().get_token_class_results()


# pylint: enable=W0613


def get_sequence_class_result(  # type: ignore
    input_ids, attention_mask, token_type_ids, boxes, model, images=None  # pylint: disable=W0613
) -> SequenceClassResult:
    """
    sequence class result
    """
    return DatapointXfund().get_sequence_class_results()


class TestHFLayoutLmTokenClassifier:
    """
    Test HFLayoutLmTokenClassifier
    """

    @staticmethod
    @mark.requires_pt
    @patch(
        "deepdoctection.extern.hflayoutlm.get_pytorch_requirement", MagicMock(return_value=("torch", False, "DUMMY"))
    )
    @patch("deepdoctection.extern.hflayoutlm.PretrainedConfig.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMForTokenClassification.from_pretrained", MagicMock())
    def test_hf_layout_lm_does_not_build_when_pt_not_available() -> None:
        """
        HFLayoutLmTokenClassifier needs pytorch. Construction fails, when requirement is not satisfied
        """

        # Arrange, Act & Assert
        with raises(ImportError):
            HFLayoutLmTokenClassifier("path/to/json", "path/to/model", ["foo"], ["B", "I", "O"])

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.PretrainedConfig.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMForTokenClassification.from_pretrained", MagicMock())
    def test_categories_are_constructed_properly() -> None:
        """
        HFLayoutLmTokenClassifier creates a full category set depending on semantics, tagging or by passing the
        set of categories directly
        """

        # Arrange, Act & Assert
        with raises(ValueError):
            HFLayoutLmTokenClassifier("path/to/json", "path/to/model", ["foo"], None)

        # Arrange
        categories_semantics = [TokenClasses.header]
        categories_bio = [BioTag.begin, BioTag.inside, BioTag.outside]

        # Act
        model = HFLayoutLmTokenClassifier("path/to/json", "path/to/model", categories_semantics, categories_bio)

        # Assert
        assert set(model.categories.values()) == {BioTag.outside, get_type("B-header"), get_type("I-header")}

        # Arrange
        categories_explicit = {"1": get_type("B-header"), "2": get_type("I-header"), "3": get_type("O")}

        # Act
        model = HFLayoutLmTokenClassifier("path/to/json", "path/to/model", categories=categories_explicit)

        # Assert
        assert model.categories == categories_explicit

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMForTokenClassification.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.PretrainedConfig.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.predict_token_classes", MagicMock(side_effect=get_token_class_results))
    def test_hf_layout_lm_predicts_token(
        layoutlm_input_for_predictor: JsonDict,
        categories_semantics: List[str],
        categories_bio: List[str],
        token_class_names: List[str],
    ) -> None:
        """
        HFLayoutLmTokenClassifier calls predict_token_classes and post processes TokenClassResult correctly
        """

        # Arrange
        categories_semantics = [TokenClasses.header]
        categories_bio = [BioTag.begin, BioTag.inside, BioTag.outside]
        layoutlm = HFLayoutLmTokenClassifier("path/to/json", "path/to/model", categories_semantics, categories_bio)
        layoutlm.model.device = "cpu"

        # Act
        inputs = {
            "image_ids": layoutlm_input_for_predictor["image_ids"],
            "width": layoutlm_input_for_predictor["width"],
            "height": layoutlm_input_for_predictor["height"],
            "ann_ids": layoutlm_input_for_predictor["ann_ids"],
            "tokens": layoutlm_input_for_predictor["tokens"],
            "bbox": torch.tensor(layoutlm_input_for_predictor["bbox"]),
            "input_ids": torch.tensor(layoutlm_input_for_predictor["input_ids"]),
            "attention_mask": torch.tensor(layoutlm_input_for_predictor["attention_mask"]),
            "token_type_ids": torch.tensor(layoutlm_input_for_predictor["token_type_ids"]),
        }

        results = layoutlm.predict(**inputs)

        # Assert
        assert len(results) == 18
        class_names = [res.class_name for res in results]
        assert class_names == token_class_names


class TestHFLayoutLmv2TokenClassifier:
    """
    Test HFLayoutLmv2TokenClassifier
    """

    @staticmethod
    @mark.requires_pt
    @patch(
        "deepdoctection.extern.hflayoutlm.get_pytorch_requirement", MagicMock(return_value=("torch", False, "DUMMY"))
    )
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2Config.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2ForTokenClassification.from_pretrained", MagicMock())
    def test_hf_layout_lm_does_not_build_when_pt_not_available() -> None:
        """
        HFLayoutLmv2TokenClassifier needs pytorch. Construction fails, when requirement is not satisfied
        """

        # Arrange, Act & Assert
        with raises(ImportError):
            HFLayoutLmv2TokenClassifier("path/to/json", "path/to/model", ["foo"], ["B", "I", "O"])

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2Config.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2ForTokenClassification.from_pretrained", MagicMock())
    def test_categories_are_constructed_properly() -> None:
        """
        HFLayoutLmv2TokenClassifier creates a full category set depending on semantics, tagging or by passing the
        set of categories directly
        """

        # Arrange, Act & Assert
        with raises(ValueError):
            HFLayoutLmv2TokenClassifier("path/to/json", "path/to/model", ["foo"], None)

        # Arrange
        categories_semantics = [TokenClasses.header]
        categories_bio = [BioTag.begin, BioTag.inside, BioTag.outside]

        # Act
        model = HFLayoutLmv2TokenClassifier("path/to/json", "path/to/model", categories_semantics, categories_bio)

        # Assert
        assert set(model.categories.values()) == {BioTag.outside, get_type("B-header"), get_type("I-header")}

        # Arrange
        categories_explicit = {"1": get_type("B-header"), "2": get_type("I-header"), "3": get_type("O")}

        # Act
        model = HFLayoutLmv2TokenClassifier("path/to/json", "path/to/model", categories=categories_explicit)

        # Assert
        assert model.categories == categories_explicit

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2ForTokenClassification.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2Config.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.predict_token_classes", MagicMock(side_effect=get_token_class_results))
    def test_hf_layout_lm_predicts_token(
        layoutlm_v2_input: JsonDict,
        categories_semantics: List[str],
        categories_bio: List[str],
        token_class_names: List[str],
    ) -> None:
        """
        HFLayoutLmTokenClassifier calls predict_token_classes and post processes TokenClassResult correctly
        """

        # Arrange
        categories_semantics = [TokenClasses.header]
        categories_bio = [BioTag.begin, BioTag.inside, BioTag.outside]
        layoutlm_v2 = HFLayoutLmv2TokenClassifier("path/to/json", "path/to/model", categories_semantics, categories_bio)
        layoutlm_v2.model.device = "cpu"

        # Act
        inputs = {
            "image_ids": layoutlm_v2_input["image_ids"],
            "width": layoutlm_v2_input["width"],
            "height": layoutlm_v2_input["height"],
            "ann_ids": layoutlm_v2_input["ann_ids"],
            "tokens": layoutlm_v2_input["tokens"],
            "bbox": torch.tensor(layoutlm_v2_input["bbox"]),
            "input_ids": torch.tensor(layoutlm_v2_input["input_ids"]),
            "attention_mask": torch.tensor(layoutlm_v2_input["attention_mask"]),
            "token_type_ids": torch.tensor(layoutlm_v2_input["token_type_ids"]),
            "image": torch.tensor(layoutlm_v2_input["image"]),
        }

        results = layoutlm_v2.predict(**inputs)

        # Assert
        assert len(results) == 18
        class_names = [res.class_name for res in results]
        assert class_names == token_class_names


class TestHFLayoutLmv3TokenClassifier:
    """
    Test HFLayoutLmv3TokenClassifier
    """

    @staticmethod
    @mark.requires_pt
    @patch(
        "deepdoctection.extern.hflayoutlm.get_pytorch_requirement", MagicMock(return_value=("torch", False, "DUMMY"))
    )
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3Config.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3ForTokenClassification.from_pretrained", MagicMock())
    def test_hf_layout_lm_does_not_build_when_pt_not_available() -> None:
        """
        HFLayoutLmv3TokenClassifier needs pytorch. Construction fails, when requirement is not satisfied
        """

        # Arrange, Act & Assert
        with raises(ImportError):
            HFLayoutLmv3TokenClassifier("path/to/json", "path/to/model", ["foo"], ["B", "I", "O"])

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3Config.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3ForTokenClassification.from_pretrained", MagicMock())
    def test_categories_are_constructed_properly() -> None:
        """
        HFLayoutLmv3TokenClassifier creates a full category set depending on semantics, tagging or by passing the
        set of categories directly
        """

        # Arrange, Act & Assert
        with raises(ValueError):
            HFLayoutLmv3TokenClassifier("path/to/json", "path/to/model", ["foo"], None)

        # Arrange
        categories_semantics = [TokenClasses.header]
        categories_bio = [BioTag.begin, BioTag.inside, BioTag.outside]

        # Act
        model = HFLayoutLmv3TokenClassifier("path/to/json", "path/to/model", categories_semantics, categories_bio)

        # Assert
        assert set(model.categories.values()) == {BioTag.outside, get_type("B-header"), get_type("I-header")}

        # Arrange
        categories_explicit = {"1": get_type("B-header"), "2": get_type("I-header"), "3": get_type("O")}

        # Act
        model = HFLayoutLmv3TokenClassifier("path/to/json", "path/to/model", categories=categories_explicit)

        # Assert
        assert model.categories == categories_explicit

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3ForTokenClassification.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3Config.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.predict_token_classes", MagicMock(side_effect=get_token_class_results))
    def test_hf_layout_lm_predicts_token(
        layoutlm_v2_input: JsonDict,
        categories_semantics: List[str],
        categories_bio: List[str],
        token_class_names: List[str],
    ) -> None:
        """
        HFLayoutLmTokenClassifier calls predict_token_classes and post processes TokenClassResult correctly
        """

        # Arrange
        categories_semantics = [TokenClasses.header]
        categories_bio = [BioTag.begin, BioTag.inside, BioTag.outside]
        layoutlm_v3 = HFLayoutLmv3TokenClassifier("path/to/json", "path/to/model", categories_semantics, categories_bio)
        layoutlm_v3.model.device = "cpu"

        # Act
        inputs = {
            "image_ids": layoutlm_v2_input["image_ids"],
            "width": layoutlm_v2_input["width"],
            "height": layoutlm_v2_input["height"],
            "ann_ids": layoutlm_v2_input["ann_ids"],
            "tokens": layoutlm_v2_input["tokens"],
            "bbox": torch.tensor(layoutlm_v2_input["bbox"]),
            "input_ids": torch.tensor(layoutlm_v2_input["input_ids"]),
            "attention_mask": torch.tensor(layoutlm_v2_input["attention_mask"]),
            "token_type_ids": torch.tensor(layoutlm_v2_input["token_type_ids"]),
            "pixel_values": torch.tensor(layoutlm_v2_input["image"]),
        }

        results = layoutlm_v3.predict(**inputs)

        # Assert
        assert len(results) == 18
        class_names = [res.class_name for res in results]
        assert class_names == token_class_names


class TestHFLayoutLmSequenceClassifier:
    """
    Test HFLayoutLmSequenceClassifier
    """

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMForSequenceClassification.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.PretrainedConfig.from_pretrained", MagicMock())
    @patch(
        "deepdoctection.extern.hflayoutlm.predict_sequence_classes", MagicMock(side_effect=get_sequence_class_result)
    )
    def test_hf_layout_lm_predicts_sequence_class(
        layoutlm_input_for_predictor: JsonDict,
    ) -> None:
        """
        HFLayoutLmTokenClassifier calls predict_sequence_classes and post processes SequenceClassResult correctly
        """

        # Arrange
        categories = {"1": get_type("FOO"), "2": get_type("BAK")}
        layoutlm = HFLayoutLmSequenceClassifier("path/to/json", "path/to/model", categories)
        layoutlm.model.device = "cpu"

        # Act
        inputs = {
            "image_ids": layoutlm_input_for_predictor["image_ids"],
            "width": layoutlm_input_for_predictor["width"],
            "height": layoutlm_input_for_predictor["height"],
            "ann_ids": layoutlm_input_for_predictor["ann_ids"],
            "tokens": layoutlm_input_for_predictor["tokens"],
            "bbox": torch.tensor(layoutlm_input_for_predictor["bbox"]),
            "input_ids": torch.tensor(layoutlm_input_for_predictor["input_ids"]),
            "attention_mask": torch.tensor(layoutlm_input_for_predictor["attention_mask"]),
            "token_type_ids": torch.tensor(layoutlm_input_for_predictor["token_type_ids"]),
        }

        results = layoutlm.predict(**inputs)

        # Assert
        assert results.class_name == "BAK"


class TestHFLayoutLmv2SequenceClassifier:
    """
    Test HFLayoutLmv2SequenceClassifier
    """

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2ForSequenceClassification.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv2Config.from_pretrained", MagicMock())
    @patch(
        "deepdoctection.extern.hflayoutlm.predict_sequence_classes", MagicMock(side_effect=get_sequence_class_result)
    )
    def test_hf_layout_lm_v2_predicts_sequence_class(
        layoutlm_v2_input: JsonDict,
    ) -> None:
        """
        HFLayoutLmv2SequenceClassifier calls predict_sequence_classes and post processes SequenceClassResult correctly
        """

        # Arrange
        categories = {"1": get_type("FOO"), "2": get_type("BAK")}
        layoutlm_v2 = HFLayoutLmv2SequenceClassifier("path/to/json", "path/to/model", categories)
        layoutlm_v2.model.device = "cpu"

        # Act
        inputs = {
            "image_ids": layoutlm_v2_input["image_ids"],
            "width": layoutlm_v2_input["width"],
            "height": layoutlm_v2_input["height"],
            "ann_ids": layoutlm_v2_input["ann_ids"],
            "tokens": layoutlm_v2_input["tokens"],
            "bbox": torch.tensor(layoutlm_v2_input["bbox"]),
            "input_ids": torch.tensor(layoutlm_v2_input["input_ids"]),
            "attention_mask": torch.tensor(layoutlm_v2_input["attention_mask"]),
            "token_type_ids": torch.tensor(layoutlm_v2_input["token_type_ids"]),
            "image": torch.tensor(layoutlm_v2_input["image"]),
        }

        results = layoutlm_v2.predict(**inputs)

        # Assert
        assert results.class_name == "BAK"


class TestHFLayoutLmv3SequenceClassifier:
    """
    Test HFLayoutLmv3SequenceClassifier
    """

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3ForSequenceClassification.from_pretrained", MagicMock())
    @patch("deepdoctection.extern.hflayoutlm.LayoutLMv3Config.from_pretrained", MagicMock())
    @patch(
        "deepdoctection.extern.hflayoutlm.predict_sequence_classes", MagicMock(side_effect=get_sequence_class_result)
    )
    def test_hf_layout_lm_v3_predicts_sequence_class(
        layoutlm_v2_input: JsonDict,
    ) -> None:
        """
        HFLayoutLmv3SequenceClassifier calls predict_sequence_classes and post processes SequenceClassResult correctly
        """

        # Arrange
        categories = {"1": get_type("FOO"), "2": get_type("BAK")}
        layoutlm_v3 = HFLayoutLmv3SequenceClassifier("path/to/json", "path/to/model", categories)
        layoutlm_v3.model.device = "cpu"

        # Act
        inputs = {
            "image_ids": layoutlm_v2_input["image_ids"],
            "width": layoutlm_v2_input["width"],
            "height": layoutlm_v2_input["height"],
            "ann_ids": layoutlm_v2_input["ann_ids"],
            "tokens": layoutlm_v2_input["tokens"],
            "bbox": torch.tensor(layoutlm_v2_input["bbox"]),
            "input_ids": torch.tensor(layoutlm_v2_input["input_ids"]),
            "attention_mask": torch.tensor(layoutlm_v2_input["attention_mask"]),
            "token_type_ids": torch.tensor(layoutlm_v2_input["token_type_ids"]),
            "pixel_values": torch.tensor(layoutlm_v2_input["image"]),
        }

        results = layoutlm_v3.predict(**inputs)

        # Assert
        assert results.class_name == "BAK"



# -*- coding: utf-8 -*-
# File: tqdm.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
tqdm related functions. (Settings, options, etc.)
"""

from typing import Dict, Optional, Union

from tqdm import tqdm

from .detection_types import TqdmType

__all__ = ["get_tqdm", "get_tqdm_default_kwargs"]


def get_tqdm_default_kwargs(
    **kwargs: Optional[Union[str, int, float]]
) -> Dict[str, Union[str, float, bool, int, None]]:
    """
    Return default arguments to be used with tqdm.
    :param kwargs: extra arguments to be used.
    """

    return dict(
        total=kwargs.get("total"),
        leave=True,
        smoothing=0.5,
        dynamic_ncols=True,
        ascii=True,
        bar_format="{l_bar}{bar}|{n_fmt}/{total_fmt}[{elapsed}<{remaining},{rate_noinv_fmt}]",
        mininterval=5,
    )


def get_tqdm(total: Optional[Union[int, float]] = None, **kwargs: Union[str, int, float]) -> TqdmType:
    """
    Get tqdm progress bar with some default options to have consistent style.

    :param total:  The number of expected iterations.
    :return: A tqdm instance
    """

    default_tqdm_setting = get_tqdm_default_kwargs(total=total)
    default_tqdm_setting.update(kwargs)

    return tqdm(**default_tqdm_setting)  # type: ignore



# -*- coding: utf-8 -*-
# File: tpdetect.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
TP Faster RCNN model as predictor for deepdoctection pipeline
"""

from copy import copy
from pathlib import Path
from typing import List, Mapping, Optional, Union

from ..utils.detection_types import ImageType, Requirement
from ..utils.file_utils import get_tensorflow_requirement, get_tensorpack_requirement, tensorpack_available
from ..utils.metacfg import set_config_by_yaml
from ..utils.settings import ObjectTypes, TypeOrStr
from .base import DetectionResult, ObjectDetector, PredictorBase

if tensorpack_available():
    from .tp.tpcompat import TensorpackPredictor
    from .tp.tpfrcnn.config.config import model_frcnn_config
    from .tp.tpfrcnn.modeling.generalized_rcnn import ResNetFPNModel
    from .tp.tpfrcnn.predict import tp_predict_image


class TPFrcnnDetector(TensorpackPredictor, ObjectDetector):
    """
    Tensorpack Faster-RCNN implementation with FPN and optional Cascade-RCNN. The backbones Resnet-50, Resnet-101 and
    their Resnext counterparts are also available. Normalization options (group normalization, synchronized batch
    normalization) for backbone in FPN can be chosen as well.

    Currently, masks are not included in the data model. However, Mask-RCNN is implemented in this version.

    There are hardly any adjustments to the original implementation of Tensorpack. As post-processing, another round
    of NMS can be carried out for the output, which operates on a class-agnostic basis. For a discussion, see
    <https://github.com/facebookresearch/detectron2/issues/978> .

        config_path = ModelCatalog.get_full_path_configs("dd/tp/conf_frcnn_rows.yaml")
        weights_path = ModelDownloadManager.maybe_download_weights_and_configs("item/model-162000.data-00000-of-00001")
        categories = ModelCatalog.get_profile("item/model-162000.data-00000-of-00001").categories

        tp_predictor = TPFrcnnDetector("tp_frcnn", config_path,weights_path,categories)  # first argument is only a name
        detection_results = tp_predictor.predict(bgr_image_np_array)

    """

    def __init__(
        self,
        path_yaml: str,
        path_weights: str,
        categories: Mapping[str, TypeOrStr],
        config_overwrite: Optional[List[str]] = None,
        ignore_mismatch: bool = False,
    ):
        """
        Set up the predictor.

        The configuration of the model is stored in a yaml-file, which needs to be passed through. For more details,
        please check

        <https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/config.py> .

        Mask-Mode could be used as well here provided the data structure is established.

        :param path_yaml: The path to the yaml config
        :param path_weights: The path to the model checkpoint
        :param categories: A dict with key (indices) and values (category names). Index 0 must be reserved for a
                           dummy 'BG' category.
        :param config_overwrite: Overwrite some hyperparameters defined by the yaml file with some new values. E.g.
                                 ["OUTPUT.FRCNN_NMS_THRESH=0.3","OUTPUT.RESULT_SCORE_THRESH=0.6"]
        :param ignore_mismatch: When True will ignore mismatches between checkpoint weights and models. This is needed
                                if a pre-trained model is to be fine-tuned on a custom dataset.
        """
        self.name = "_".join(Path(path_weights).parts[-3:])
        self.path_yaml = path_yaml
        self.categories = copy(categories)  # type: ignore
        self.config_overwrite = config_overwrite
        model = TPFrcnnDetector.set_model(path_yaml, self.categories, config_overwrite)
        super().__init__(model, path_weights, ignore_mismatch)
        assert self._number_gpus > 0, "Model only support inference with GPU"

    @staticmethod
    def set_model(
        path_yaml: str, categories: Mapping[str, ObjectTypes], config_overwrite: Union[List[str], None]
    ) -> ResNetFPNModel:
        """
        Calls all necessary methods to build TP ResNetFPNModel

        :param path_yaml: path to the model config
        :param categories: A dict of categories with indices as keys
        :param config_overwrite: A list with special config attributes reset (consult  __init__)
        :return: The FPNResnet model.
        """

        if config_overwrite is None:
            config_overwrite = []

        hyper_param_config = set_config_by_yaml(path_yaml)

        if len(config_overwrite):
            hyper_param_config.update_args(config_overwrite)

        model_frcnn_config(config=hyper_param_config, categories=categories, print_summary=False)
        return ResNetFPNModel(config=hyper_param_config)

    def predict(self, np_img: ImageType) -> List[DetectionResult]:
        """
        Prediction per image.

        :param np_img: image as numpy array
        :return: A list of DetectionResult
        """
        detection_results = tp_predict_image(
            np_img,
            self.tp_predictor,
            self._model.cfg.PREPROC.SHORT_EDGE_SIZE,
            self._model.cfg.PREPROC.MAX_SIZE,
            self._model.cfg.MRCNN.ACCURATE_PASTE,
        )
        return self._map_category_names(detection_results)

    def _map_category_names(self, detection_results: List[DetectionResult]) -> List[DetectionResult]:
        """
        Populating category names to detection results

        :param detection_results: list of detection results
        :return: List of detection results with attribute class_name populated
        """
        for result in detection_results:
            result.class_name = self._model.cfg.DATA.CLASS_DICT[str(result.class_id)]
        return detection_results

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_tensorflow_requirement(), get_tensorpack_requirement()]

    def clone(self) -> PredictorBase:
        return self.__class__(
            self.path_yaml, self.path_weights, self.categories, self.config_overwrite, self.ignore_mismatch
        )



# -*- coding: utf-8 -*-
# File: test_pubtabnet.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Testing module datasets.instances.pubtabnet
"""

from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.datasets import Pubtabnet

from ...test_utils import collect_datapoint_from_dataflow, get_test_path
from .conftest import get_white_image


@mark.basic
def test_dataset_pubtabnet_returns_image() -> None:
    """
    test dataset pubtabnet returns image
    """

    # Arrange
    pubtabnet = Pubtabnet()
    pubtabnet.dataflow.get_workdir = get_test_path  # type: ignore
    pubtabnet.dataflow.annotation_files = {"all": "test_file.jsonl"}
    df = pubtabnet.dataflow.build()

    # Act
    df_list = collect_datapoint_from_dataflow(df)
    assert len(df_list) == 3


@mark.basic
@patch("deepdoctection.mapper.pubstruct.load_image_from_file", MagicMock(side_effect=get_white_image))
def test_dataset_pubtabnet_with_load_image_returns_image() -> None:
    """
    test dataset publaynet returns image
    """

    # Arrange
    pubtabnet = Pubtabnet()
    pubtabnet.dataflow.get_workdir = get_test_path  # type: ignore
    pubtabnet.dataflow.annotation_files = {"all": "test_file.jsonl"}
    df = pubtabnet.dataflow.build(load_image=True)

    # Act
    df_list = collect_datapoint_from_dataflow(df)
    assert len(df_list) == 3



# -*- coding: utf-8 -*-
# File: context.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some useful contextmanagers for various tasks
"""

import subprocess
from contextlib import contextmanager
from errno import ENOENT
from glob import iglob
from os import path, remove
from tempfile import NamedTemporaryFile
from time import perf_counter as timer
from typing import Any, Generator, Iterator, Optional, Tuple, Union

import numpy as np
from cv2 import imwrite

from .detection_types import ImageType
from .logger import logger

__all__ = ["timeout_manager", "save_tmp_file", "timed_operation"]


@contextmanager
def timeout_manager(proc, seconds: Optional[int] = None) -> Iterator[str]:  # type: ignore
    """
    Manager for time handling while some process being called

       with timeout_manager(some_process,60) as timeout:
           ...

    :param proc: process
    :param seconds: seconds to wait

    """
    try:
        if not seconds:
            yield proc.communicate()[1]
            return

        try:
            _, error_string = proc.communicate(timeout=seconds)
            yield error_string
        except subprocess.TimeoutExpired:
            proc.terminate()
            proc.kill()
            proc.returncode = -1
            raise RuntimeError("Tesseract process timeout")  # pylint: disable=W0707
    finally:
        if proc.stdin is not None:
            proc.stdin.close()
        if proc.stdout is not None:
            proc.stdout.close()
        if proc.stderr is not None:
            proc.stderr.close()


@contextmanager
def save_tmp_file(image: Union[str, ImageType, bytes], prefix: str) -> Iterator[Tuple[str, str]]:
    """
    Save image temporarily and handle the clean-up once not necessary anymore

        with save_tmp_file(some_np_image,"tmp") as (tmp_name, input_file_name):
            ....

    :param image: image as string or numpy array
    :param prefix: prefix of the temp file name
    """
    try:
        with NamedTemporaryFile(prefix=prefix, delete=False) as file:
            if isinstance(image, str):
                yield file.name, path.realpath(path.normpath(path.normcase(image)))
                return
            if isinstance(image, (np.ndarray, np.generic)):
                input_file_name = file.name + ".PNG"
                imwrite(input_file_name, image)
                yield file.name, input_file_name
            if isinstance(image, bytes):
                input_file_name = file.name
                file.write(image)
                file.flush()
                yield file.name, input_file_name
    finally:
        for file_name in iglob(file.name + "*" if file.name else file.name):
            try:
                remove(file_name)
            except OSError as error:
                if error.errno != ENOENT:
                    raise error


# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")


@contextmanager
def timed_operation(message: str, log_start: bool = False) -> Generator[Any, None, None]:
    """
    Contextmanager with a timer. Can therefore be used in a with statement.

    :param message: a log to print
    :param log_start: whether to print also the beginning
    """

    assert len(message)
    if log_start:
        logger.info("start task: %s ...", message)
    start = timer()
    yield
    logger.info("%s total: %s sec.", message, round(timer() - start, 4))



# -*- coding: utf-8 -*-
# File: info.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for storing dataset info (e.g. general meta data or categories)
"""

from copy import copy
from dataclasses import dataclass, field
from itertools import chain
from typing import Any, Dict, List, Literal, Mapping, Optional, Sequence, Set, Union, overload

from ..utils.settings import DefaultType, ObjectTypes, TypeOrStr, get_type
from ..utils.utils import call_only_once

__all__ = ["DatasetInfo", "DatasetCategories", "get_merged_categories"]


@overload
def _get_dict(l: Sequence[ObjectTypes], name_as_key: Literal[True], starts_with: int = ...) -> Dict[ObjectTypes, str]:
    ...


@overload
def _get_dict(l: Sequence[ObjectTypes], name_as_key: Literal[False], starts_with: int = ...) -> Dict[str, ObjectTypes]:
    ...


@overload
def _get_dict(
    l: Sequence[ObjectTypes], name_as_key: bool, starts_with: int = ...
) -> Union[Dict[ObjectTypes, str], Dict[str, ObjectTypes]]:
    ...


def _get_dict(
    l: Sequence[ObjectTypes], name_as_key: bool, starts_with: int = 1
) -> Union[Dict[ObjectTypes, str], Dict[str, ObjectTypes]]:
    """
    Converts a list into a dict, where keys/values are the list indices.

    :param l: A list of categories
    :param name_as_key: Whether to return the dict with category names as key (True)
    :param starts_with: index count start
    :return: A dictionary of list indices/list elements.
    """
    if name_as_key:
        return {v: str(k) for k, v in enumerate(l, starts_with)}
    return {str(k): v for k, v in enumerate(l, starts_with)}


@dataclass
class DatasetInfo:
    """
    DatasetInfo is a simple dataclass that stores some meta-data information about a dataset.

    `name`: Name of the dataset. Using the name you can retrieve the dataset from the
    `registry.DatasetRegistry`.

    `description`: Short description of the dataset.

    `license`: License to the dataset.

    `url`: url, where the dataset can be downloaded from.

    `splits`: A dict of splits. The value must store the relative path, where the split can be found.

    `type`: The type describes whether this is a dataset for object detection (pass 'OBJECT_DETECTION'),
    sequence classification (pass 'SEQUENCE_CLASSIFICATION') or token classification ('TOKEN_CLASSIFICATION').
    Optionally, pass `None`.
    """

    name: str
    description: str = field(default="")
    license: str = field(default="")
    url: Union[str, Sequence[str]] = field(default="")
    splits: Mapping[str, str] = field(default_factory=dict)
    type: ObjectTypes = field(default=DefaultType.default_type)

    def get_split(self, key: str) -> str:
        """
        Get the split directory by its key (if it exists).

        :param key: The key to a split (i.e. "train", "val", "test")
        :return: The local directory path to the split. An empty string if the key doesn't exist.
        """

        return self.splits[key]


@dataclass
class DatasetCategories:
    """
    Categories and their sub-categories are managed in this separate class. Since categories can be filtered or
    sub-categories can be swapped with categories, the list of all categories must be adapted at the same time
    after replacement/filtering. DatasetCategories manages these transformations. The class is also responsible
    for the index/category name relationship and guarantees that a sequence of natural numbers for the categories
    is always returned as the category-id even after replacing and/or filtering.

    `init_categories`: A list of category names. The list must include all categories that can occur within the
    annotations.

    `init_sub_categories`: A dict of categories/sub-categories. Each sub-category that can appear in the
    annotations in combination with a category must be listed.

    **Example:**

        An annotation file hast the category/sub-category combinations for three datapoints:

            (cat1,s1),(cat1,s2), (cat2,s2).

        You must list `init_categories`, `init_sub_categories` as follows:

            init_categories = [cat1,cat2]
            init_sub_categories = {cat1: [s1,s2],cat2: [s2]}

    Use `filter_categories` or `set_cat_to_sub_cat` to filter or swap categories with sub-categories.
    """

    init_categories: Sequence[ObjectTypes] = field(default_factory=list)
    init_sub_categories: Mapping[ObjectTypes, Mapping[ObjectTypes, Sequence[ObjectTypes]]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self._categories_update = self.init_categories
        self._cat_to_sub_cat: Optional[Mapping[ObjectTypes, ObjectTypes]] = None
        self._allow_update = True
        self._init_sanity_check()

    @overload
    def get_categories(
        self, *, name_as_key: Literal[True], init: bool = ..., filtered: bool = ...
    ) -> Mapping[ObjectTypes, str]:
        ...

    @overload
    def get_categories(
        self, *, name_as_key: Literal[False] = ..., init: bool = ..., filtered: bool = ...
    ) -> Mapping[str, ObjectTypes]:
        ...

    @overload
    def get_categories(
        self, as_dict: Literal[False], name_as_key: bool = False, init: bool = False, filtered: bool = False
    ) -> Sequence[ObjectTypes]:
        ...

    @overload
    def get_categories(
        self, as_dict: Literal[True] = ..., name_as_key: bool = False, init: bool = False, filtered: bool = False
    ) -> Union[Mapping[ObjectTypes, str], Mapping[str, ObjectTypes]]:
        ...

    def get_categories(
        self, as_dict: bool = True, name_as_key: bool = False, init: bool = False, filtered: bool = False
    ) -> Union[Sequence[ObjectTypes], Mapping[ObjectTypes, str], Mapping[str, ObjectTypes]]:
        """
        Get categories of a dataset. The returned value also respects modifications of the inventory like filtered
        categories of replaced categories with sub categories. However, you must correctly pass arguments to return the
        state you want.

        :param as_dict: Will pass a dict if set to 'True' otherwise a list.
        :param name_as_key: Categories are stored as key/value pair in a dict with integers as keys. name_as_key set to
                            "False" will swap keys and values.
        :param init: If set to "True" it will return the list/dict of categories as initially provided. Manipulations
                     due to replacing/filtering will not be regarded.
        :param filtered: If set to "True" will return an unfiltered list of all categories. If a replacing has been
                         invoked selected sub categories will be returned.
        :return: A dict of index/category names (or the other way around) or a list of category names.
        """
        if init:
            if as_dict:
                return _get_dict(self.init_categories, name_as_key)
            return list(copy(self.init_categories))
        if filtered:
            if as_dict:
                if hasattr(self, "_categories_filter_update"):
                    return _get_dict(self._categories_filter_update, name_as_key)
                return _get_dict(self._categories_update, name_as_key)
            if hasattr(self, "_categories_filter_update"):
                return self._categories_filter_update
            return list(copy(self._categories_update))
        if as_dict:
            return _get_dict(self._categories_update, name_as_key)
        return list(copy(self._categories_update))

    def get_sub_categories(
        self,
        categories: Optional[Union[TypeOrStr, Sequence[TypeOrStr]]] = None,
        sub_categories: Optional[Mapping[TypeOrStr, Union[TypeOrStr, Sequence[TypeOrStr]]]] = None,
        keys: bool = True,
        values_as_dict: bool = True,
        name_as_key: bool = False,
    ) -> Mapping[ObjectTypes, Any]:
        """
        Returns a dict of list with a category name and their sub categories.

        :param categories: A single category or list of category names
        :param sub_categories: A mapping of categories to sub category keys on which the result should be filtered. Only
                               relevant, if `keys=False`
        :param keys: Will only pass keys if set to `True`.
        :param values_as_dict: Will generate a dict with indices and sub category value names if set to `True`.
        :param name_as_key: sub category values are stored as key/value pair in a dict with integers as keys.
                            name_as_key set to `False` will swap keys and values.
        :return: Dict with all selected categories.
        """
        _categories: Sequence[ObjectTypes]
        if isinstance(categories, (ObjectTypes, str)):
            _categories = [get_type(categories)]
        elif isinstance(categories, list):
            _categories = [get_type(category) for category in categories]
        else:
            _categories = self.get_categories(as_dict=False, filtered=True)
            if _categories is None:
                _categories = []
        if sub_categories is None:
            sub_categories = {}

        sub_cat: Dict[ObjectTypes, Union[ObjectTypes, List[ObjectTypes]]] = {}
        for cat in _categories:  # pylint: disable=R1702
            assert cat in self.get_categories(  # pylint: disable=E1135
                as_dict=False, filtered=True
            ), f"{cat} not in categories. Maybe it has been replaced with sub category"
            sub_cat_dict = self.init_sub_categories.get(cat)
            if sub_cat_dict is None:
                if self._cat_to_sub_cat:
                    for key, val in self._cat_to_sub_cat.items():
                        possible_sub_cats = self.init_sub_categories[key]
                        for pos in possible_sub_cats:
                            if pos == val:
                                sub_cat_dict_key = self.init_sub_categories[key]
                                sub_cat_list = sub_cat_dict_key.get(cat)
                                if sub_cat_list is None:
                                    sub_cat[cat] = []
                                else:
                                    sub_cat[cat] = list(copy(sub_cat_list))
            else:
                sub_cat[cat] = list(sub_cat_dict.keys())
        if not keys:
            sub_cat_values = {}
            for category, value in sub_cat.items():
                if category not in sub_categories:
                    continue
                sub_cat_tmp: Dict[str, Union[Dict[str, str], Sequence[str]]] = {}
                sub_categories_list: Union[
                    ObjectTypes, str, List[Sequence[Union[ObjectTypes, str]]], Sequence[Union[ObjectTypes, str]]
                ]
                if isinstance(sub_categories[category], ObjectTypes):
                    sub_categories_list = [sub_categories[category]]
                else:
                    sub_categories_list = sub_categories[category]
                for sub_cat_key in value:
                    if sub_cat_key not in sub_categories_list:
                        continue
                    if values_as_dict:
                        if not name_as_key:
                            sub_cat_tmp[sub_cat_key] = {
                                str(k): v
                                for k, v in enumerate(self.init_sub_categories[category][get_type(sub_cat_key)], 1)
                            }
                        else:
                            sub_cat_tmp[sub_cat_key] = {
                                v: str(k)
                                for k, v in enumerate(self.init_sub_categories[category][get_type(sub_cat_key)], 1)
                            }
                    else:
                        sub_cat_tmp[sub_cat_key] = self.init_sub_categories[category][get_type(sub_cat_key)]
                sub_cat_values[category] = sub_cat_tmp
            return sub_cat_values

        return sub_cat

    @call_only_once
    def set_cat_to_sub_cat(self, cat_to_sub_cat: Dict[TypeOrStr, TypeOrStr]) -> None:
        """
        Change category representation if sub-categories are available. Pass a dictionary of the main category
        and the requested sub-category. This will change the dictionary of categories and the category names
        and category ids of the image annotation in the dataflow. Using more than one sub category per category will
        not make sense.

        This method can only be called once per object. Re-setting or further replacing of categories would make the
        code messy and is therefore not allowed.

        **Example:**

                  cat_to_sub_cat={cat1: sub_cat1}

            will replace cat1 with sub_cat1 as category. This will also be respected when returning datapoints.

        :param cat_to_sub_cat: A dict of pairs of category/sub-category. Note that the combination must be available
                               according to the initial settings.
        """

        _cat_to_sub_cat = {get_type(key): get_type(value) for key, value in cat_to_sub_cat.items()}
        if not self._allow_update:
            raise PermissionError("Replacing categories with sub categories is not allowed")
        self._categories_update = self.init_categories
        categories = self.get_categories(name_as_key=True)
        cats_or_sub_cats = [
            self.init_sub_categories.get(cat, {cat: [cat]}).get(_cat_to_sub_cat.get(cat, cat), [cat])
            for cat in categories  # pylint: disable=E1133
        ]
        self._cat_to_sub_cat = _cat_to_sub_cat
        _categories_update_list = list(chain(*cats_or_sub_cats))

        # we must keep the order of _categories_update_list if the elements are unique
        if len(set(_categories_update_list)) != len(_categories_update_list):
            self._categories_update = list(set(_categories_update_list))
        else:
            self._categories_update = _categories_update_list

    @call_only_once
    def filter_categories(self, categories: Union[TypeOrStr, List[TypeOrStr]]) -> None:
        """
        Filter categories of a dataset. This will keep all the categories chosen and remove all others.
        This method can only be called once per object.

        :param categories: A single category name or a list of category names.
        """

        if not self._allow_update:
            raise PermissionError("Filtering categories is not allowed")
        if isinstance(categories, (ObjectTypes, str)):
            categories = [get_type(categories)]
        else:
            categories = [get_type(category) for category in categories]

        self._categories_filter_update: Sequence[ObjectTypes] = [
            cat for cat in self._categories_update if cat in categories
        ]

    @property
    def cat_to_sub_cat(self) -> Optional[Mapping[ObjectTypes, ObjectTypes]]:
        """
        cat_to_sub_cat
        """
        return self._cat_to_sub_cat

    def is_cat_to_sub_cat(self) -> bool:
        """
        returns `True` if a category is replaced with sub categories
        """
        if self._cat_to_sub_cat is not None:
            return True
        return False

    def is_filtered(self) -> bool:
        """
        return `True` if categories are filtered
        """
        if hasattr(self, "_categories_filter_update"):
            return True
        return False

    def _init_sanity_check(self) -> None:
        # all values of possible sub categories must be listed
        assert all(isinstance(val, dict) for val in self.init_sub_categories.values())


def get_merged_categories(*categories: DatasetCategories) -> DatasetCategories:
    """
    Given a set of `DatasetCategories`, a `DatasetCategories` instance will be returned that summarize the category
    properties of merged dataset. This means it will save the union of all possible categories as its init categories.
    Regarding sub categories, only those will be accessible if and only if they are a sub category of a category for all
    merged datasets. E.g. if dataset A has category `foo` with sub category `foo`:`bak` and dataset B has category `foo`
    as well but no sub category than the merged dataset will have no sub categories at all. Whereas in a similar setting
    dataset B has sub category `foo`:`bak`, then `bak` will be an optional sub category for the merged dataset as well.

    :param categories: A tuple/list of dataset categories
    :return: An instance of `DatasetCategories` to be used as `DatasetCategories` for merged datasets
    """

    # working with lists is not possible as the order of categories is important here
    init_categories = []
    categories_update = []
    categories_filtered = []
    for cat in categories:
        for label in cat.init_categories:
            if label not in init_categories:
                init_categories.append(label)
        for label in cat.get_categories(as_dict=False):
            if label not in categories_update:
                categories_update.append(label)
        for label in cat.get_categories(as_dict=False, filtered=True):
            if label not in categories_filtered:
                categories_filtered.append(label)

    # select categories with sub categories. Only categories that appear in this list can be candidates for having
    # sub categories in the merged dataset
    intersect_sub_cat_keys = set(categories[0].init_sub_categories.keys()).intersection(
        *[cat.init_sub_categories.keys() for cat in categories[1:]]
    )
    intersect_init_sub_cat = {}
    for key in intersect_sub_cat_keys:
        # select all sub categories from all datasets for a given key
        sub_cat_per_key = [cat.init_sub_categories[key] for cat in categories]
        # select only sub categories that appear in all datasets
        intersect_sub_cat_per_key = set(sub_cat_per_key[0].keys()).intersection(
            *[sub_cats.keys() for sub_cats in sub_cat_per_key[1:]]
        )
        # form a set of possible sub category values. To get a list of all values from all dataset, take the union
        intersect_init_sub_cat_values = {}
        for sub_cat_key in intersect_sub_cat_per_key:
            val: Set[ObjectTypes] = set()
            for cat in categories:
                val.update(cat.init_sub_categories[key][sub_cat_key])
            intersect_init_sub_cat_values[sub_cat_key] = list(val)
        intersect_init_sub_cat[key] = intersect_init_sub_cat_values

    # Next, build the DatasetCategories instance.
    merged_categories = DatasetCategories(init_categories=init_categories, init_sub_categories=intersect_init_sub_cat)
    merged_categories._categories_update = categories_update  # pylint: disable = W0212
    merged_categories._allow_update = False  # pylint: disable = W0212
    setattr(merged_categories, "_categories_filter_update", categories_filtered)
    return merged_categories



# -*- coding: utf-8 -*-
# File: test_common.py
#
# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.common
"""
from pytest import mark

from deepdoctection.datapoint import Image
from deepdoctection.pipe import AnnotationNmsService, MatchingService
from deepdoctection.utils.settings import LayoutType, Relationships


class TestMatchingService:
    """
    Test MatchingService
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._parent_categories = LayoutType.cell
        self._child_categories = LayoutType.word
        self._matching_rule = "ioa"
        self._iou_threshold = None
        self._ioa_threshold = 0.499

        self.matching_service = MatchingService(
            self._parent_categories,
            self._child_categories,
            self._matching_rule,  # type: ignore
            self._iou_threshold if self._matching_rule in ["iou"] else self._ioa_threshold,  # type: ignore
        )

    @mark.basic
    def test_integration_pipeline_component(self, dp_image_fully_segmented_unrelated_words: Image) -> None:
        """
        test matching service assigns child annotation according to matching rule correctly to parental
        annotations
        """

        dp = dp_image_fully_segmented_unrelated_words

        # Act
        dp = self.matching_service.pass_datapoint(dp)

        parent_anns = dp.get_annotation(category_names=self._parent_categories)
        child_anns = dp.get_annotation(category_names=self._child_categories)

        relationships_word_first_parent = parent_anns[0].get_relationship(Relationships.child)
        relationships_word_third_parent = parent_anns[2].get_relationship(Relationships.child)

        # Assert
        assert len(relationships_word_first_parent) == 1
        assert len(relationships_word_third_parent) == 1

        assert relationships_word_first_parent[0] == child_anns[0].annotation_id
        assert relationships_word_third_parent[0] == child_anns[1].annotation_id


class TestAnnotationNmsService:
    """
    Test AnnotationNmsService
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._nms_pairs = [["row", "column"]]
        self._thresholds = [0.01]
        self._categories = ["row", "column"]

        self.nms_service = AnnotationNmsService(self._nms_pairs, self._thresholds)

    @mark.requires_pt
    def test_integration_pipeline_component(self, dp_image_fully_segmented_unrelated_words: Image) -> None:
        """
        test annotation nms service suppresses the annotations within groups
        """

        dp = dp_image_fully_segmented_unrelated_words

        # Act
        dp = self.nms_service.pass_datapoint(dp)

        anns = dp.get_annotation(category_names=self._categories)

        # Assert
        assert len(anns) == 2



# -*- coding: utf-8 -*-
# File: tokenclass.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for token classification pipeline
"""

from copy import copy
from typing import Any, List, Literal, Optional, Sequence, Union

from ..datapoint.image import Image
from ..extern.hflayoutlm import HFLayoutLmSequenceClassifierBase, HFLayoutLmTokenClassifierBase
from ..mapper.laylmstruct import image_to_layoutlm_features
from ..utils.detection_types import JsonDict
from ..utils.file_utils import transformers_available
from ..utils.settings import BioTag, LayoutType, ObjectTypes, PageType, TokenClasses, WordType
from .base import LanguageModelPipelineComponent
from .registry import pipeline_component_registry

if transformers_available():
    from transformers import LayoutLMTokenizerFast, RobertaTokenizerFast, XLMRobertaTokenizerFast

    _ARCHITECTURES_TO_TOKENIZER = {
        ("LayoutLMForTokenClassification", False): LayoutLMTokenizerFast.from_pretrained(
            "microsoft/layoutlm-base-uncased"
        ),
        ("LayoutLMForSequenceClassification", False): LayoutLMTokenizerFast.from_pretrained(
            "microsoft/layoutlm-base-uncased"
        ),
        ("LayoutLMv2ForTokenClassification", False): LayoutLMTokenizerFast.from_pretrained(
            "microsoft/layoutlm-base-uncased"
        ),
        ("LayoutLMv2ForSequenceClassification", False): LayoutLMTokenizerFast.from_pretrained(
            "microsoft/layoutlm-base-uncased"
        ),
        ("LayoutLMv2ForTokenClassification", True): XLMRobertaTokenizerFast.from_pretrained("xlm-roberta-base"),
        ("LayoutLMv2ForSequenceClassification", True): XLMRobertaTokenizerFast.from_pretrained("xlm-roberta-base"),
        ("LayoutLMv3ForSequenceClassification", False): RobertaTokenizerFast.from_pretrained(
            "roberta-base", add_prefix_space=True
        ),
        ("LayoutLMv3ForTokenClassification", False): RobertaTokenizerFast.from_pretrained(
            "roberta-base", add_prefix_space=True
        ),
    }


def get_tokenizer_from_architecture(architecture_name: str, use_xlm_tokenizer: bool) -> Any:
    """
    We do not use the tokenizer for a particular model that the transformer library provides. Thie mapping therefore
    returns the tokenizer that should be used for a particular model.

    :param architecture_name: The model as stated in the transformer library.
    :param use_xlm_tokenizer: True if one uses the LayoutXLM. (The model cannot be distinguished from LayoutLMv2).
    :return: Tokenizer instance to use.
    """
    return _ARCHITECTURES_TO_TOKENIZER[(architecture_name, use_xlm_tokenizer)]


@pipeline_component_registry.register("LMTokenClassifierService")
class LMTokenClassifierService(LanguageModelPipelineComponent):
    """
    Pipeline component for token classification

    **Example**

            # setting up compulsory ocr service
            tesseract_config_path = ModelCatalog.get_full_path_configs("/dd/conf_tesseract.yaml")
            tess = TesseractOcrDetector(tesseract_config_path)
            ocr_service = TextExtractionService(tess)

            # hf tokenizer and token classifier
            tokenizer = LayoutLMTokenizerFast.from_pretrained("microsoft/layoutlm-base-uncased")
            layoutlm = HFLayoutLmTokenClassifier(categories= ['B-answer', 'B-header', 'B-question', 'E-answer',
                                                               'E-header', 'E-question', 'I-answer', 'I-header',
                                                               'I-question', 'O', 'S-answer', 'S-header', 'S-question'])

            # token classification service
            layoutlm_service = LMTokenClassifierService(tokenizer,layoutlm)

            pipe = DoctectionPipe(pipeline_component_list=[ocr_service,layoutlm_service])

            path = "path/to/some/form"
            df = pipe.analyze(path=path)

            for dp in df:
                ...
    """

    def __init__(
        self,
        tokenizer: Any,
        language_model: HFLayoutLmTokenClassifierBase,
        padding: Literal["max_length", "do_not_pad", "longest"] = "max_length",
        truncation: bool = True,
        return_overflowing_tokens: bool = False,
        use_other_as_default_category: bool = False,
        segment_positions: Optional[Union[LayoutType, Sequence[LayoutType]]] = None,
        sliding_window_stride: int = 0,
    ) -> None:
        """
        :param tokenizer: Token classifier, typing allows currently anything. This will be changed in the future
        :param language_model: language model token classifier
        :param padding: A padding strategy to be passed to the tokenizer. Must bei either `max_length, longest` or
                        `do_not_pad`.
        :param truncation: If "True" will truncate to a maximum length specified with the argument max_length or to the
                           maximum acceptable input length for the model if that argument is not provided. This will
                           truncate token by token, removing a token from the longest sequence in the pair if a pair of
                           sequences (or a batch of pairs) is provided.
                           If `False` then no truncation (i.e., can output batch with sequence lengths greater than the
                           model maximum admissible input size).
        :param return_overflowing_tokens: If a sequence (due to a truncation strategy) overflows the overflowing tokens
                           can be returned as an additional batch element. Not that in this case, the number of input
                           batch samples will be smaller than the output batch samples.
        :param use_other_as_default_category: When predicting token classes, it might be possible that some words might
                                              not get sent to the model because they are categorized as not eligible
                                              token (e.g. empty string). If set to `True` it will assign all words
                                              without token the `BioTag.outside` token.
        :param segment_positions: Using bounding boxes of segment instead of words improves model accuracy significantly
                              for models that have been trained on segments rather than words.
                              Choose a single or a sequence of layout segments to use their bounding boxes. Note, that
                              the layout segments need to have a child-relationship with words. If a word does not
                              appear as child, it will use the word bounding box.
        :param sliding_window_stride: If the output of the tokenizer exceeds the max_length sequence length, a sliding
                              windows will be created with each window having max_length sequence input. When using
                              `sliding_window_stride=0` no strides will be created, otherwise it will create slides
                              with windows shifted `sliding_window_stride` to the right.
        """
        self.language_model = language_model
        self.padding = padding
        self.truncation = truncation
        self.return_overflowing_tokens = return_overflowing_tokens
        self.use_other_as_default_category = use_other_as_default_category
        self.segment_positions = segment_positions
        self.sliding_window_stride = sliding_window_stride
        if self.use_other_as_default_category:
            categories_name_as_key = {val: key for key, val in self.language_model.categories.items()}
            self.default_key: ObjectTypes
            if BioTag.outside in categories_name_as_key:
                self.default_key = BioTag.outside
            else:
                self.default_key = TokenClasses.other
            self.other_name_as_key = {self.default_key: categories_name_as_key[self.default_key]}
        super().__init__(self._get_name(), tokenizer, image_to_layoutlm_features)
        self.required_kwargs = {
            "tokenizer": self.tokenizer,
            "padding": self.padding,
            "truncation": self.truncation,
            "return_overflowing_tokens": self.return_overflowing_tokens,
            "return_tensors": "pt",
            "segment_positions": self.segment_positions,
            "sliding_window_stride": self.sliding_window_stride,
        }
        self.required_kwargs.update(self.language_model.default_kwargs_for_input_mapping())
        self._init_sanity_checks()

    def serve(self, dp: Image) -> None:
        lm_input = self.mapping_to_lm_input_func(**self.required_kwargs)(dp)
        if lm_input is None:
            return
        lm_output = self.language_model.predict(**lm_input)

        # turn to word level predictions and remove all special tokens
        lm_output = [
            token
            for token in lm_output
            if token.token_id
            not in [self.tokenizer.cls_token_id, self.tokenizer.sep_token_id, self.tokenizer.pad_token_id]
            and not token.token.startswith("##")
        ]

        words_populated: List[str] = []
        for token in lm_output:
            if token.uuid not in words_populated:
                if token.class_name == token.semantic_name:
                    token_class_name_id = token.class_id
                else:
                    token_class_name_id = None
                self.dp_manager.set_category_annotation(
                    token.semantic_name, token_class_name_id, WordType.token_class, token.uuid
                )
                self.dp_manager.set_category_annotation(token.bio_tag, None, WordType.tag, token.uuid)
                self.dp_manager.set_category_annotation(
                    token.class_name, token.class_id, WordType.token_tag, token.uuid
                )
                words_populated.append(token.uuid)

        if self.use_other_as_default_category:
            word_anns = dp.get_annotation(LayoutType.word)
            for word in word_anns:
                if WordType.token_class not in word.sub_categories:
                    self.dp_manager.set_category_annotation(
                        TokenClasses.other,
                        self.other_name_as_key[self.default_key],
                        WordType.token_class,
                        word.annotation_id,
                    )
                if WordType.tag not in word.sub_categories:
                    self.dp_manager.set_category_annotation(BioTag.outside, None, WordType.tag, word.annotation_id)
                if WordType.token_tag not in word.sub_categories:
                    self.dp_manager.set_category_annotation(
                        self.default_key,
                        self.other_name_as_key[self.default_key],
                        WordType.token_tag,
                        word.annotation_id,
                    )

    def clone(self) -> "LMTokenClassifierService":
        return self.__class__(
            copy(self.tokenizer),
            self.language_model.clone(),
            self.padding,
            self.truncation,
            self.return_overflowing_tokens,
            self.use_other_as_default_category,
            self.segment_positions,
            self.sliding_window_stride,
        )

    def get_meta_annotation(self) -> JsonDict:
        return dict(
            [
                ("image_annotations", []),
                ("sub_categories", {LayoutType.word: {WordType.token_class, WordType.tag, WordType.token_tag}}),
                ("relationships", {}),
                ("summaries", []),
            ]
        )

    def _get_name(self) -> str:
        return f"lm_token_class_{self.language_model.name}"

    def _init_sanity_checks(self) -> None:
        tokenizer_class = self.language_model.model.config.tokenizer_class
        use_xlm_tokenizer = False
        if tokenizer_class is not None:
            use_xlm_tokenizer = True
        tokenizer_reference = get_tokenizer_from_architecture(
            self.language_model.model.__class__.__name__, use_xlm_tokenizer
        )
        if not isinstance(self.tokenizer, type(tokenizer_reference)):
            raise ValueError(
                f"You want to use {type(self.tokenizer)} but you should use {type(tokenizer_reference)} "
                f"in this framework"
            )


@pipeline_component_registry.register("LMSequenceClassifierService")
class LMSequenceClassifierService(LanguageModelPipelineComponent):
    """
    Pipeline component for sequence classification

    **Example**

            # setting up compulsory ocr service
            tesseract_config_path = ModelCatalog.get_full_path_configs("/dd/conf_tesseract.yaml")
            tess = TesseractOcrDetector(tesseract_config_path)
            ocr_service = TextExtractionService(tess)

            # hf tokenizer and token classifier
            tokenizer = LayoutLMTokenizerFast.from_pretrained("microsoft/layoutlm-base-uncased")
            layoutlm = HFLayoutLmSequenceClassifier("path/to/config.json","path/to/model.bin",
                                                     categories=["handwritten", "presentation", "resume"])

            # token classification service
            layoutlm_service = LMSequenceClassifierService(tokenizer,layoutlm)

            pipe = DoctectionPipe(pipeline_component_list=[ocr_service,layoutlm_service])

            path = "path/to/some/form"
            df = pipe.analyze(path=path)

            for dp in df:
                ...

    """

    def __init__(
        self,
        tokenizer: Any,
        language_model: HFLayoutLmSequenceClassifierBase,
        padding: Literal["max_length", "do_not_pad", "longest"] = "max_length",
        truncation: bool = True,
        return_overflowing_tokens: bool = False,
    ) -> None:
        """
        :param tokenizer: Tokenizer, typing allows currently anything. This will be changed in the future
        :param language_model: language model sequence classifier
        :param padding: A padding strategy to be passed to the tokenizer. Must bei either `max_length, longest` or
                        `do_not_pad`.
        :param truncation: If "True" will truncate to a maximum length specified with the argument max_length or to the
                           maximum acceptable input length for the model if that argument is not provided. This will
                           truncate token by token, removing a token from the longest sequence in the pair if a pair of
                           sequences (or a batch of pairs) is provided.
                           If `False` then no truncation (i.e., can output batch with sequence lengths greater than the
                           model maximum admissible input size).
        :param return_overflowing_tokens: If a sequence (due to a truncation strategy) overflows the overflowing tokens
                           can be returned as an additional batch element. Not that in this case, the number of input
                           batch samples will be smaller than the output batch samples.
        """
        self.language_model = language_model
        self.padding = padding
        self.truncation = truncation
        self.return_overflowing_tokens = return_overflowing_tokens
        super().__init__(self._get_name(), tokenizer, image_to_layoutlm_features)
        self.required_kwargs = {
            "tokenizer": self.tokenizer,
            "padding": self.padding,
            "truncation": self.truncation,
            "return_overflowing_tokens": self.return_overflowing_tokens,
            "return_tensors": "pt",
        }
        self.required_kwargs.update(self.language_model.default_kwargs_for_input_mapping())
        self._init_sanity_checks()

    def serve(self, dp: Image) -> None:
        lm_input = self.mapping_to_lm_input_func(**self.required_kwargs)(dp)
        if lm_input is None:
            return
        lm_output = self.language_model.predict(**lm_input)
        self.dp_manager.set_summary_annotation(
            PageType.document_type, lm_output.class_name, lm_output.class_id, None, lm_output.score
        )

    def clone(self) -> "LMSequenceClassifierService":
        return self.__class__(
            copy(self.tokenizer),
            self.language_model.clone(),
            self.padding,
            self.truncation,
            self.return_overflowing_tokens,
        )

    def get_meta_annotation(self) -> JsonDict:
        return dict(
            [
                ("image_annotations", []),
                ("sub_categories", {}),
                ("relationships", {}),
                ("summaries", [PageType.document_type]),
            ]
        )

    def _get_name(self) -> str:
        return f"lm_sequence_class_{self.language_model.name}"

    def _init_sanity_checks(self) -> None:
        tokenizer_class = self.language_model.model.config.tokenizer_class
        use_xlm_tokenizer = False
        if tokenizer_class is not None:
            use_xlm_tokenizer = True
        tokenizer_reference = get_tokenizer_from_architecture(
            self.language_model.model.__class__.__name__, use_xlm_tokenizer
        )
        if not isinstance(self.tokenizer, type(tokenizer_reference)):
            raise ValueError(
                f"You want to use {type(self.tokenizer)} but you should use {type(tokenizer_reference)} "
                f"in this framework"
            )



# -*- coding: utf-8 -*-
# File: __init__.py

"""
Init file for deepdoctection package
"""

import os
import sys
from typing import TYPE_CHECKING

from packaging import version

from .utils.file_utils import _LazyModule, get_tf_version, pytorch_available, tf_available
from .utils.logger import logger

__version__ = 0.22

_IMPORT_STRUCTURE = {
    "analyzer": ["get_dd_analyzer", "build_analyzer", "load_page", "load_document"],
    "configs": [],
    "dataflow": [
        "DataFlowTerminated",
        "DataFlowResetStateNotCalled",
        "DataFlowReentrantGuard",
        "DataFlow",
        "RNGDataFlow",
        "ProxyDataFlow",
        "TestDataSpeed",
        "FlattenData",
        "MapData",
        "MapDataComponent",
        "RepeatedData",
        "ConcatData",
        "JoinData",
        "CacheData",
        "CustomDataFromList",
        "CustomDataFromIterable",
        "SerializerJsonlines",
        "SerializerTabsepFiles",
        "SerializerFiles",
        "CocoParser",
        "SerializerCoco",
        "SerializerPdfDoc",
        "MultiThreadMapData",
        "MultiProcessMapData",
        "DataFromList",
        "DataFromIterable",
        "FakeData",
        "PickleSerializer",
        "MeanFromDataFlow",
        "StdFromDataFlow",
    ],
    "datapoint": [
        "ann_from_dict",
        "Annotation",
        "CategoryAnnotation",
        "ImageAnnotation",
        "SummaryAnnotation",
        "ContainerAnnotation",
        "coco_iou",
        "area",
        "intersection",
        "np_iou",
        "iou",
        "BoundingBoxError",
        "BoundingBox",
        "intersection_box",
        "crop_box_from_image",
        "local_to_global_coords",
        "global_to_local_coords",
        "merge_boxes",
        "rescale_coords",
        "intersection_boxes",
        "convert_b64_to_np_array",
        "convert_np_array_to_b64",
        "convert_np_array_to_b64_b",
        "convert_pdf_bytes_to_np_array_v2",
        "box_to_point4",
        "point4_to_box",
        "as_dict",
        "ImageAnnotationBaseView",
        "Image",
        "Word",
        "Layout",
        "Cell",
        "Table",
        "Page",
    ],
    "datasets": [
        "DatasetAdapter",
        "DatasetBase",
        "MergeDataset",
        "CustomDataset",
        "DataFlowBaseBuilder",
        "DatasetInfo",
        "DatasetCategories",
        "get_merged_categories",
        "dataset_registry",
        "get_dataset",
        "print_dataset_infos",
        "dataflow_to_json",
    ],
    "datasets.instances": [
        "DocLayNet",
        "DocLayNetSeq",
        "Fintabnet",
        "Funsd",
        "IIITar13K",
        "LayoutTest",
        "Publaynet",
        "Pubtables1MDet",
        "Pubtables1MStruct",
        "Pubtabnet",
        "Rvlcdip",
        "Xfund",
    ],
    "datasets.instances.xsl": [],
    "eval": [
        "AccuracyMetric",
        "ConfusionMetric",
        "PrecisionMetric",
        "RecallMetric",
        "F1Metric",
        "PrecisionMetricMicro",
        "RecallMetricMicro",
        "F1MetricMicro",
        "MetricBase",
        "CocoMetric",
        "Evaluator",
        "metric_registry",
        "get_metric",
        "TableTree",
        "CustomConfig",
        "TEDS",
        "TedsMetric",
        "EvalCallback",
    ],
    "extern": [
        "PredictorBase",
        "DetectionResult",
        "ObjectDetector",
        "PdfMiner",
        "TextRecognizer",
        "TokenClassResult",
        "SequenceClassResult",
        "LMTokenClassifier",
        "LMSequenceClassifier",
        "LanguageDetector",
        "ImageTransformer",
        "InferenceResize",
        "D2FrcnnDetector",
        "Jdeskewer",
        "DoctrTextlineDetector",
        "DoctrTextRecognizer",
        "FasttextLangDetector",
        "HFDetrDerivedDetector",
        "HFLayoutLmTokenClassifierBase",
        "HFLayoutLmTokenClassifier",
        "HFLayoutLmv2TokenClassifier",
        "HFLayoutLmSequenceClassifier",
        "HFLayoutLmv2SequenceClassifier",
        "HFLayoutLmv3SequenceClassifier",
        "ModelProfile",
        "ModelCatalog",
        "print_model_infos",
        "ModelDownloadManager",
        "PdfPlumberTextDetector",
        "TesseractOcrDetector",
        "TextractOcrDetector",
        "TPFrcnnDetector",
    ],
    "extern.pt": ["set_torch_auto_device", "get_num_gpu"],
    "extern.tp": ["disable_tfv2", "ModelDescWithConfig", "TensorpackPredictor"],
    "extern.tp.tpfrcnn": ["CustomResize", "anchors_and_labels", "augment"],
    "extern.tp.tpfrcnn.utils": ["area", "pairwise_intersection", "pairwise_iou"],
    "extern.tp.tpfrcnn.config": ["model_frcnn_config", "train_frcnn_config"],
    "extern.tp.tpfrcnn.modeling": ["ResNetFPNModel"],
    "mapper": [
        "cat_to_sub_cat",
        "re_assign_cat_ids",
        "filter_cat",
        "filter_summary",
        "image_to_cat_id",
        "remove_cats",
        "add_summary",
        "coco_to_image",
        "image_to_coco",
        "image_to_d2_frcnn_training",
        "pt_nms_image_annotations",
        "to_wandb_image",
        "image_to_hf_detr_training",
        "DetrDataCollator",
        "image_to_layoutlm",
        "image_to_raw_layoutlm_features",
        "raw_features_to_layoutlm_features",
        "LayoutLMDataCollator",
        "image_to_layoutlm_features",
        "DataCollator",
        "LayoutLMFeatures",
        "MappingContextManager",
        "DefaultMapper",
        "maybe_get_fake_score",
        "LabelSummarizer",
        "curry",
        "match_anns_by_intersection",
        "to_image",
        "maybe_load_image",
        "maybe_remove_image",
        "maybe_remove_image_from_category",
        "image_ann_to_image",
        "maybe_ann_to_sub_image",
        "xml_to_dict",
        "to_page",
        "page_dict_to_page",
        "pascal_voc_dict_to_image",
        "prodigy_to_image",
        "image_to_prodigy",
        "pub_to_image_uncur",
        "pub_to_image",
        "image_to_tp_frcnn_training",
        "tf_nms_image_annotations",
        "xfund_to_image",
    ],
    "pipe": [
        "DatapointManager",
        "PipelineComponent",
        "PredictorPipelineComponent",
        "LanguageModelPipelineComponent",
        "ImageTransformPipelineComponent",
        "Pipeline",
        "DetectResultGenerator",
        "SubImageLayoutService",
        "ImageCroppingService",
        "MatchingService",
        "PageParsingService",
        "AnnotationNmsService",
        "MultiThreadPipelineComponent",
        "DoctectionPipe",
        "LanguageDetectionService",
        "ImageLayoutService",
        "get_tokenizer_from_architecture",
        "LMTokenClassifierService",
        "LMSequenceClassifierService",
        "TableSegmentationRefinementService",
        "generate_html_string",
        "pipeline_component_registry",
        "TableSegmentationService",
        "SegmentationResult",
        "TextExtractionService",
        "TextOrderService",
        "SimpleTransformPipelineComponent",
    ],
    "train": [
        "D2Trainer",
        "train_d2_faster_rcnn",
        "LayoutLMTrainer",
        "train_hf_layoutlm",
        "train_faster_rcnn",
        "DetrDerivedTrainer",
        "train_hf_detr",
    ],
    "utils": [
        "timeout_manager",
        "save_tmp_file",
        "timed_operation",
        "get_tensorflow_requirement",
        "tf_addons_available",
        "get_tf_addons_requirements",
        "tensorpack_available",
        "get_tensorpack_requirement",
        "get_pytorch_requirement",
        "lxml_available",
        "get_lxml_requirement",
        "apted_available",
        "get_apted_requirement",
        "distance_available",
        "get_distance_requirement",
        "transformers_available",
        "get_transformers_requirement",
        "detectron2_available",
        "get_detectron2_requirement",
        "tesseract_available",
        "set_tesseract_path",
        "get_tesseract_version",
        "get_tesseract_requirement",
        "pdf_to_ppm_available",
        "pdf_to_cairo_available",
        "get_poppler_requirement",
        "pdfplumber_available",
        "get_pdfplumber_requirement",
        "cocotools_available",
        "get_cocotools_requirement",
        "scipy_available",
        "sklearn_available",
        "get_sklearn_requirement",
        "qpdf_available",
        "boto3_available",
        "get_boto3_requirement",
        "aws_available",
        "get_aws_requirement",
        "doctr_available",
        "get_doctr_requirement",
        "fasttext_available",
        "get_fasttext_requirement",
        "wandb_available",
        "get_wandb_requirement",
        "load_image_from_file",
        "load_bytes_from_pdf_file",
        "get_load_image_func",
        "maybe_path_or_pdf",
        "download",
        "mkdir_p",
        "is_file_extension",
        "load_json",
        "FileExtensionError",
        "is_uuid_like",
        "get_uuid_from_str",
        "get_uuid",
        "logger",
        "set_logger_dir",
        "auto_set_dir",
        "get_logger_dir",
        "AttrDict",
        "set_config_by_yaml",
        "save_config_to_yaml",
        "config_to_cli_str",
        "decrypt_pdf_document",
        "get_pdf_file_reader",
        "get_pdf_file_writer",
        "PDFStreamer",
        "pdf_to_np_array",
        "ObjectTypes",
        "TypeOrStr",
        "object_types_registry",
        "DefaultType",
        "PageType",
        "SummaryType",
        "DocumentType",
        "LayoutType",
        "TableType",
        "CellType",
        "WordType",
        "TokenClasses",
        "BioTag",
        "TokenClassWithTag",
        "Relationships",
        "Languages",
        "DatasetType",
        "update_all_types_dict",
        "get_type",
        "sub_path",
        "get_package_path",
        "get_configs_dir_path",
        "get_weights_dir_path",
        "get_dataset_dir_path",
        "get_tqdm",
        "get_tqdm_default_kwargs",
        "ResizeTransform",
        "InferenceResize",
        "normalize_image",
        "pad_image",
        "PadTransform",
        "delete_keys_from_dict",
        "split_string",
        "string_to_dict",
        "to_bool",
        "call_only_once",
        "get_rng",
        "FileExtensionError",
        "is_file_extension",
        "draw_text",
        "draw_boxes",
        "interactive_imshow",
    ],
}

if not tf_available() and not pytorch_available():
    logger.info(
        "Neither Tensorflow or Pytorch are available. You will not be able to use any Deep Learning model from"
        "the library."
    )

# disable TF warnings for versions > 2.4.1
if tf_available():
    if version.parse(get_tf_version()) > version.parse("2.4.1"):
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
    try:
        import tensorflow.python.util.deprecation as deprecation  # type: ignore # pylint: disable=E0401,R0402

        deprecation._PRINT_DEPRECATION_WARNINGS = False  # pylint: disable=W0212
    except Exception:  # pylint: disable=W0703
        try:
            from tensorflow.python.util import deprecation  # type: ignore # pylint: disable=E0401

            deprecation._PRINT_DEPRECATION_WARNINGS = False  # pylint: disable=W0212
        except Exception:  # pylint: disable=W0703
            pass

# Direct imports for type-checking
if TYPE_CHECKING:
    from .analyzer import *
    from .dataflow import *
    from .datapoint import *
    from .datasets import *
    from .eval import *
    from .extern import *
    from .mapper import *
    from .pipe import *
    from .train import *
    from .utils import *

else:
    sys.modules[__name__] = _LazyModule(
        __name__,
        globals()["__file__"],
        _IMPORT_STRUCTURE,
        module_spec=__spec__,
        extra_objects={"__version__": __version__},
    )



# -*- coding: utf-8 -*-
# File: common.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/common.py>
"""


import cv2
import numpy as np
from tensorpack.dataflow.imgaug import ImageAugmentor, ResizeTransform  # pylint: disable=E0401

from ....utils.file_utils import cocotools_available

if cocotools_available():
    import pycocotools.mask as coco_mask


class CustomResize(ImageAugmentor):
    """
    Try resizing the shortest edge to a certain number while avoiding the longest edge to exceed max_size.
    """

    def __init__(self, short_edge_length, max_size, interp=cv2.INTER_LINEAR):
        """
        :param short_edge_length: a [min, max] interval from which to sample the shortest edge length.
        :param max_size: maximum allowed longest edge length.
        :param interp: cv2 interpolation mode
        """
        super().__init__()
        if isinstance(short_edge_length, int):
            short_edge_length = (short_edge_length, short_edge_length)
        self.short_edge_length = short_edge_length
        self.max_size = max_size
        self.interp = interp

    def get_transform(self, img):
        """
        get transform
        """
        h, w = img.shape[:2]
        size = self.rng.randint(self.short_edge_length[0], self.short_edge_length[1] + 1)
        scale = size * 1.0 / min(h, w)
        if h < w:
            new_h, new_w = size, scale * w
        else:
            new_h, new_w = scale * h, size
        if max(new_h, new_w) > self.max_size:
            scale = self.max_size * 1.0 / max(new_h, new_w)
            new_h = new_h * scale
            new_w = new_w * scale
        new_w = int(new_w + 0.5)
        new_h = int(new_h + 0.5)
        return ResizeTransform(h, w, new_h, new_w, self.interp)


def polygons_to_mask(polys, height, width, intersect=False):
    """
    Convert polygons to binary masks.

    :param polys: a list of nx2 float array. Each array contains many (x, y) coordinates.#
    :param intersect: intersect
    :param width: width
    :param height: height
    :return: A binary matrix of (height, width)
    """

    polys = [p.flatten().tolist() for p in polys]
    assert len(polys) > 0, "Polygons are empty!"

    rles = coco_mask.frPyObjects(polys, height, width)
    rle = coco_mask.merge(rles, intersect)
    return coco_mask.decode(rle)


def clip_boxes(boxes, shape):
    """
    :param boxes: (...)x4, float
    :param shape: h, w
    """

    orig_shape = boxes.shape
    boxes = boxes.reshape([-1, 4])
    h, w = shape
    boxes[:, [0, 1]] = np.maximum(boxes[:, [0, 1]], 0)
    boxes[:, 2] = np.minimum(boxes[:, 2], w)
    boxes[:, 3] = np.minimum(boxes[:, 3], h)
    return boxes.reshape(orig_shape)


def filter_boxes_inside_shape(boxes, shape):
    """
    :param boxes: (nx4), float
    :param shape: (h, w)

    :return: indices: (k, )
            selection: (kx4)
    """

    assert boxes.ndim == 2, boxes.shape
    assert len(shape) == 2, shape
    h, w = shape
    indices = np.where((boxes[:, 0] >= 0) & (boxes[:, 1] >= 0) & (boxes[:, 2] <= w) & (boxes[:, 3] <= h))[0]
    return indices, boxes[indices, :]


# Much faster than utils/np_box_ops
def np_iou(box_a, box_b):
    """
    np iou
    """

    def to_xywh(box):
        box = box.copy()
        box[:, 2] -= box[:, 0]
        box[:, 3] -= box[:, 1]
        return box

    ret = coco_mask.iou(to_xywh(box_a), to_xywh(box_b), np.zeros((len(box_b),), dtype=bool))
    # can accelerate even more, if using float32
    return ret.astype("float32")



# -*- coding: utf-8 -*-
# File: data.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some data samples in a separate module
"""
from deepdoctection.extern.base import DetectionResult
from deepdoctection.utils.detection_types import JsonDict
from deepdoctection.utils.settings import CellType, LayoutType

PDF_BYTES = (
    b"%PDF-1.3\n1 0 obj\n<<\n/Type /Pages\n/Count 1\n/Kids [ 3 0 R ]\n>>\nendobj\n2 0 obj\n<<\n/Producer "
    b"(PyPDF2)\n>>\nendobj\n3 0 obj\n<<\n/Type /Page\n/Parent 1 0 R\n/Resources <<\n/Font "
    b"<<\n/F1 5 0 R\n>>\n/ProcSet 6 0 R\n>>\n/MediaBox [ 0 0 612 792 ]\n/Contents 7 0 R\n>>\nendobj\n4 "
    b"0 obj\n<<\n/Type /Catalog\n/Pages 1 0 R\n>>\nendobj\n5 0 obj\n<<\n/Type /Font\n/Subtype /Type1\n/Name"
    b" /F1\n/BaseFont /Helvetica\n/Encoding /WinAnsiEncoding\n>>\nendobj\n6 0 obj\n[ /PDF /Text"
    b" ]\nendobj\n7 0 obj\n<<\n/Length 1074\n>>\nstream\n2 J\r\nBT\r\n0 0 0 rg\r\n/F1 0027 Tf\r\n57.3750"
    b" 722.2800 Td\r\n( A Simple PDF File ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 688.6080 Td\r\n("
    b" This is a small demonstration .pdf file - ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 664.7040"
    b" Td\r\n( just for use in the Virtual Mechanics tutorials. More text. And more ) Tj\r\nET\r\nBT\r\n/F1"
    b" 0010 Tf\r\n69.2500 652.7520 Td\r\n( text. And more text. And more text. And more text. )"
    b" Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 628.8480 Td\r\n( And more text. And more text."
    b" And more text. And more text. And more ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 616.8960"
    b" Td\r\n( text. And more text. Boring, zzzzz. And more text. And more text. And )"
    b" Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 604.9440 Td\r\n( more text. And more text. And more"
    b" text. And more text. And more text. ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 592.9920 Td\r\n("
    b" And more text. And more text. ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 569.0880 Td\r\n( And"
    b" more text. And more text. And more text. And more text. And more ) Tj\r\nET\r\nBT\r\n/F1 0010"
    b" Tf\r\n69.2500 557.1360 Td\r\n( text. And more text. And more text. Even more. Continued on page"
    b" 2 ...) Tj\r\nET\r\n\nendstream\nendobj\nxref\n0 8\n0000000000 65535 f \n0000000009 00000 n"
    b" \n0000000068 00000 n \n0000000108 00000 n \n0000000251 00000 n \n0000000300 00000 n \n0000000407"
    b" 00000 n \n0000000437 00000 n \ntrailer\n<<\n/Size 8\n/Root 4 0 R\n/Info 2 0"
    b" R\n>>\nstartxref\n1563\n%%EOF\n"
)

PDF_BYTES_2 = (
    b"%PDF-1.3\n1 0 obj\n<<\n/Type /Pages\n/Count 1\n/Kids [ 3 0 R ]\n>>\nendobj\n2 0 obj\n<<\n/Producer"
    b" (PyPDF2)\n>>\nendobj\n3 0 obj\n<<\n/Type /Page\n/Parent 1 0 R\n/Resources <<\n/Font <<\n/F1 5 0"
    b" R\n>>\n/ProcSet 6 0 R\n>>\n/MediaBox [ 0 0 1112 1792 ]\n/Contents 7 0 R\n>>\nendobj\n4 0"
    b" obj\n<<\n/Type /Catalog\n/Pages 1 0 R\n>>\nendobj\n5 0 obj\n<<\n/Type /Font\n/Subtype /Type1\n/Name"
    b" /F1\n/BaseFont /Helvetica\n/Encoding /WinAnsiEncoding\n>>\nendobj\n6 0 obj\n[ /PDF /Text ]\nendobj\n7"
    b" 0 obj\n<<\n/Length 676\n>>\nstream\n2 J\r\nBT\r\n0 0 0 rg\r\n/F1 0027 Tf\r\n57.3750 722.2800 Td\r\n("
    b" Simple PDF File 2 ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 688.6080 Td\r\n( ...continued from page"
    b" 1. Yet more text. And more text. And more text. ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 676.6560"
    b" Td\r\n( And more text. And more text. And more text. And more text. And more ) Tj\r\nET\r\nBT\r\n/F1"
    b" 0010 Tf\r\n69.2500 664.7040 Td\r\n( text. Oh, how boring typing this stuff. But not as boring as"
    b" watching ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 652.7520 Td\r\n( paint dry. And more text. And"
    b" more text. And more text. And more text. ) Tj\r\nET\r\nBT\r\n/F1 0010 Tf\r\n69.2500 640.8000"
    b" Td\r\n( Boring.  More, a little more text. The end, and just as well. ) Tj\r\nET\r\n\nendstream"
    b"\nendobj\nxref\n0 8\n0000000000 65535 f \n0000000009 00000 n \n0000000068 00000 n \n0000000108 00000"
    b" n \n0000000251 00000 n \n0000000300 00000 n \n0000000407 00000 n \n0000000437 00000 n \ntrailer"
    b"\n<<\n/Size 8\n/Root 4 0 R\n/Info 2 0 R\n>>\nstartxref\n1164\n%%EOF\n"
)

WORD_RESULTS = [
    DetectionResult(
        box=[10.0, 10.0, 24.0, 23.0],
        score=0.8,
        text="foo",
        block="1",
        line="2",
        class_id=1,
        class_name=LayoutType.word,
    ),
    DetectionResult(
        box=[30.0, 10.0, 38.0, 24.0],
        score=0.2,
        text="bak",
        block="1",
        line="2",
        class_id=1,
        class_name=LayoutType.word,
    ),
]


def get_detr_categories() -> JsonDict:
    """detr_categories"""
    return {
        "1": LayoutType.table,
        "2": LayoutType.column,
        "3": LayoutType.row,
        "4": CellType.column_header,
        "5": CellType.projected_row_header,
        "6": CellType.spanning,
    }
